from torch import Tensor
import torch


def points_inside_polygon_edges(points: Tensor, edges: Tensor) -> Tensor:
    """A point is inside a polygon if it is inside all the edges' half-planes.
    B sets of N points (B, N, 2), B M-sided polygon edges (B, M, 2, 2) -> (B, N, 2)
    """
    inside_points = points.clone()
    p1, p2 = edges[:, :, 0, :], edges[:, :, 1, :]
    a = (p2[:, :, 1] - p1[:, :, 1]).unsqueeze(1)  # (B, 1, M)
    b = (p1[:, :, 0] - p2[:, :, 0]).unsqueeze(1)
    c = (p2[:, :, 0] * p1[:, :, 1] - p1[:, :, 0] * p2[:, :, 1]).unsqueeze(1)
    points = points.unsqueeze(2)  # (B, N, 1, 2)
    mask = (a * points[..., 0] + b * points[..., 1] + c < 0).all(dim=2)
    inside_points[~mask] = torch.nan
    return inside_points


def segment_intersections(segments1: Tensor, segments2: Tensor) -> Tensor:
    """segments1 and segments2 have shape (N, 2, 2)"""
    A1, A2, B1, B2 = segments1[:, 0], segments1[:, 1], segments2[:, 0], segments2[:, 1]
    V1, V2 = A2 - A1, B2 - B1  # Direction vectors
    det = torch.det(torch.stack([V1.unsqueeze(1), V2.unsqueeze(1)], dim=2))
    parallel_mask = torch.abs(det) < 1e-5  # parallel or collinear
    B1_A1 = B1 - A1
    s = torch.det(torch.stack([B1_A1.unsqueeze(1), V2.unsqueeze(1)], dim=2)) / det
    t = torch.det(torch.stack([B1_A1.unsqueeze(1), V1.unsqueeze(1)], dim=2)) / det
    valid_mask = (~parallel_mask) & ((s >= 0) & (s <= 1)) & ((t >= 0) & (t <= 1))
    intersection_points = A1 + s * V1
    intersection_points[~valid_mask.squeeze()] = torch.nan
    return intersection_points


def polygon_intersection(polygons1: Tensor, polygons2: Tensor) -> Tensor:
    """the pairwise intersection polygons of 2 sets of convex n-gons.
    arguments' shape: (B, n, 2).
    the intersection polygon is made up of intersection points and inside points.
    """
    assert polygons1.shape == polygons2.shape
    num_polygons, num_sides, _ = polygons1.shape
    edges1 = torch.stack([polygons1, torch.roll(polygons1, shifts=1, dims=1)], dim=2)
    edges2 = torch.stack([polygons2, torch.roll(polygons2, shifts=1, dims=1)], dim=2)
    intersections = torch.cat(  # (B, N*N + 2N, 2)
        [
            segment_intersections(
                edges1.repeat_interleave(num_sides, dim=1).reshape(-1, 2, 2),
                edges2.repeat((1, num_sides, 1, 1)).reshape(-1, 2, 2),
            ).reshape(num_polygons, num_sides * num_sides, 2),
            points_inside_polygon_edges(points=polygons1, edges=edges2),
            points_inside_polygon_edges(points=polygons2, edges=edges1),
        ],
        dim=1,
    )
    return intersections


def polygon_area(polygons: Tensor) -> Tensor:
    """Area of polygons (B, n, 2) using shoelace formula"""
    # polygons = reorder_clockwise(polygons)
    valid_mask = ~torch.isnan(polygons).any(dim=-1)
    x, y = polygons[..., 0], polygons[..., 1]
    shift_x = torch.roll(x, shifts=-1, dims=1)
    shift_x[shift_x.isnan()] = shift_x[..., -1:].expand_as(shift_x)[shift_x.isnan()]
    shift_y = torch.roll(y, shifts=-1, dims=1)
    shift_y[shift_y.isnan()] = shift_y[..., -1:].expand_as(shift_y)[shift_y.isnan()]
    x[~valid_mask] = 0
    y[~valid_mask] = 0
    shift_x[~valid_mask] = 0
    shift_y[~valid_mask] = 0
    cross_product = x * shift_y - y * shift_x
    cross_product[~valid_mask] = 0
    area = 0.5 * torch.abs(torch.sum(cross_product, dim=1))
    return area


def polygon_iou(polygons1: Tensor, polygons2: Tensor) -> float:
    """Return intersection-over-union (Jaccard index) between two sets of polygons.

    Both sets of polygons are expected to be in ``[[x, y], ...]`` format

    Args:
        polygons1 (Tensor[N, n, 2]): first set of polygons
        polygons2 (Tensor[M, n, 2]): second set of polygons

    Returns:
        Tensor[N, M]: the NxM matrix containing the pairwise IoU values for every element in polygons1 and polygons2
    """
    num_polygons1, num_polygons2 = polygons1.shape[0], polygons2.shape[0]
    polygons1 = polygons1.repeat_interleave(num_polygons2, dim=0)
    polygons2 = polygons2.repeat((num_polygons1, 1, 1))
    intersections = polygon_area(polygon_intersection(polygons1, polygons2))
    unions = polygon_area(polygons1) + polygon_area(polygons2) - intersections
    return (intersections / unions).reshape((num_polygons1, num_polygons2))




from pathlib import Path
from typing import List, Tuple, Dict

from torch import Tensor
import numpy as np
import onnx
import onnxruntime
import pytest
import torch

from sihl.heads import ObjectDetection


BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH = 4, 256, 128, 128
NUM_CLASSES, MIN_LEVEL, MAX_LEVEL = 100, 3, 7
ONNX_VERSION, ONNX_FILE_NAME = 18, "object_detection.onnx"


@pytest.fixture()
def model() -> ObjectDetection:
    return ObjectDetection(
        in_channels=(3,) + (NUM_CHANNELS,) * MAX_LEVEL, num_classes=NUM_CLASSES
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [torch.randn((BATCH_SIZE, 3, HEIGHT, WIDTH))] + [
        torch.randn((BATCH_SIZE, NUM_CHANNELS, HEIGHT // 2**_, WIDTH // 2**_))
        for _ in range(MAX_LEVEL)
    ]


@pytest.fixture()
def targets() -> Dict[str, List[Tensor]]:
    num_objects = range(BATCH_SIZE)  # tests target with 0 objects too
    return {
        "classes": [
            torch.randint(0, NUM_CLASSES, (num_objects[_],)) for _ in range(BATCH_SIZE)
        ],
        "boxes": [
            torch.randint(0, HEIGHT, (num_objects[_], 4), dtype=torch.float)
            for _ in range(BATCH_SIZE)
        ],
    }


def test_forward(model: ObjectDetection, backbone_output: List[Tensor]) -> None:
    num_instances, scores, classes, boxes = model(backbone_output)
    assert tuple(num_instances.shape) == (BATCH_SIZE,)
    assert tuple(scores.shape) == (BATCH_SIZE, model.max_instances)
    assert tuple(classes.shape) == (BATCH_SIZE, model.max_instances)
    assert tuple(boxes.shape) == (BATCH_SIZE, model.max_instances, 4)


def test_training_step(
    model: ObjectDetection,
    backbone_output: List[Tensor],
    targets: Tuple[List[Tensor], List[Tensor]],
) -> None:
    loss, _ = model.training_step(backbone_output, **targets)
    assert loss.item() >= 0


def test_validation_step(
    model: ObjectDetection,
    backbone_output: List[Tensor],
    targets: Tuple[List[Tensor], List[Tensor]],
) -> None:
    model.on_validation_start()
    loss, _ = model.validation_step(backbone_output, **targets)
    assert loss.item() >= 0
    metrics = model.on_validation_end()
    assert metrics


@pytest.fixture()
def onnx_model(model: ObjectDetection, backbone_output: List[Tensor]) -> None:
    # TODO: upgrade to dynamo_export
    # torch.onnx.dynamo_export(
    #     model,
    #     backbone_output,
    #     export_options=torch.onnx.ExportOptions(dynamic_shapes=True),
    # ).save(ONNX_FILE_NAME)

    torch.onnx.export(
        model.to(torch.float32),
        args=(backbone_output,),
        f=ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
        output_names=[f"head0/{name}" for name, shape in model.output_shapes.items()],
        dynamic_axes=dict(
            {
                f"input_level_{lvl}": {
                    0: "batch_size",
                    2: f"height/{2**lvl}",
                    3: f"width/{2**lvl}",
                }
                for lvl in range(len(backbone_output))
            },
            **{
                f"head0/{name}": {
                    shape_idx: str(shape_value)
                    for shape_idx, shape_value in enumerate(shape)
                }
                for name, shape in model.output_shapes.items()
            },
        ),
        external_data=False,
        verify=True,
        # dynamo=True,
        # report=True,
    )
    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(
    onnx_model, model: ObjectDetection, backbone_output: List[Tensor]
) -> None:
    model.eval()
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    # just check that 99% of values are equal.
    pytorch_output = [_.detach().numpy() for _ in model(backbone_output)]
    onnx_output = onnx_session.run(None, onnx_input)
    for i in range(len(pytorch_output)):
        assert (
            np.sum(np.abs(pytorch_output[i] - onnx_output[i]) > 1) / onnx_output[i].size
            < 0.01
        )



from typing import Tuple, List, Dict

from einops import rearrange, repeat, reduce
from torch import nn, Tensor
from torch.nn import functional
import torch

from sihl.utils import coordinate_grid, polygon_iou

from .object_detection import ObjectDetection


class QuadrilateralDetection(ObjectDetection):
    """Quadrilateral detection is like object detection, expect objects are associated
    with convex quadrilaterals instead of axis-aligned rectangles."""

    def __init__(
        self,
        in_channels: List[int],
        num_classes: int,
        bottom_level: int = 3,
        top_level: int = 7,
        num_channels: int = 256,
        num_layers: int = 4,
        max_instances: int = 100,
        t_min: float = 0.2,
        t_max: float = 0.6,
        topk: int = 7,
        soft_label_decay_steps: int = 1,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible object categories.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            max_instances (int, optional): Maximum number of instances to predict in a sample. Defaults to 100.
            t_min (float, optional): Lower bound of O2F parameter t. Defaults to 0.2.
            t_max (float, optional): Upper bound of O2F parameter t. Defaults to 0.6.
            topk (int, optional): How many anchors to match to each ground truth object when copmuting the loss. Defaults to 7.
            soft_label_decay_steps (int, optional): How many training steps to perform before the one-to-few matching becomes one-to-one. Defaults to 1.
        """
        super().__init__(
            in_channels=in_channels,
            num_classes=num_classes,
            bottom_level=bottom_level,
            top_level=top_level,
            num_channels=num_channels,
            num_layers=num_layers,
            max_instances=max_instances,
            t_min=t_min,
            t_max=t_max,
            topk=topk,
            soft_label_decay_steps=soft_label_decay_steps,
        )

        del self.box_head
        self.quad_head = nn.Conv2d(self.num_channels, 8, kernel_size=1)
        self.get_boxes = self.get_quads
        self.box_iou = polygon_iou
        self.inside_box = inside_polygon
        self.box_loss = quad_loss
        self.output_shapes = {
            "num_instances": ("batch_size",),
            "scores": ("batch_size", self.max_instances),
            "classes": ("batch_size", self.max_instances),
            "quadrilaterals": ("batch_size", self.max_instances, 4, 2),
        }

    def get_quads(self, features: List[Tensor]) -> Tensor:
        quads, batch_size = [], features[0].shape[0]
        for level, x in zip(self.levels, features):
            stride, (h, w) = 2**level, x.shape[2:]
            biases = coordinate_grid(h, w, h * stride, w * stride).to(x)
            biases = repeat(biases, "c h w -> (4 c) h w", c=2)
            biases = rearrange(biases, "c h w -> 1 (h w) c", c=8)
            level_quads = rearrange(self.quad_head(x), "n c h w -> n (h w) c", c=8)
            quads.append(biases + level_quads * stride)
        quads = torch.cat(quads, dim=1)
        return convexify(quads.reshape(-1, 4, 2)).reshape(batch_size, -1, 4, 2)

    def training_step(
        self,
        inputs: List[Tensor],
        classes: List[Tensor],
        quads: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, float]]:
        loss, metrics = super().training_step(
            inputs=inputs, classes=classes, boxes=quads, is_validating=is_validating
        )
        metrics["quad_loss"] = metrics["box_loss"]
        del metrics["box_loss"]
        return loss, metrics

    def validation_step(
        self, inputs: List[Tensor], classes: List[Tensor], quads: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        num_instances, scores, pred_classes, pred_quads = self.forward(inputs)
        self.map_computer.to(pred_quads.device).update(
            [
                {"scores": s, "labels": c, "boxes": b}
                for s, c, b in zip(scores, pred_classes, quads_to_boxes(pred_quads))
            ],
            [{"labels": c, "boxes": quads_to_boxes(q)} for c, q in zip(classes, quads)],
        )
        loss, metrics = self.training_step(inputs, classes, quads, is_validating=True)
        self.loss_computer.to(loss.device).update(loss)
        return loss, metrics


def quads_to_boxes(quads):
    x, y = quads[..., 0], quads[..., 1]
    return torch.stack([x.min(-1)[0], y.min(-1)[0], x.max(-1)[0], y.max(-1)[0]], dim=-1)


def line_equation(p1: Tensor, p2: Tensor) -> Tuple[float, float, float]:
    """find (a, b, c) such that ax + by + c = 0"""
    a, b, c = p2[1] - p1[1], p1[0] - p2[0], p2[0] * p1[1] - p1[0] * p2[1]
    return a, b, c


def inside_polygon(points: Tensor, polygon: Tensor) -> Tensor:
    """A point is inside a polygon if it is inside all edges' half-planes.
    N points (N, 2), 1 M-sided polygon (M, 2) -> (N,) bool
    """
    m = polygon.shape[0]
    edges = [line_equation(polygon[i], polygon[(i + 1) % m]) for i in range(m)]
    inside = [a * points[:, 0] + b * points[:, 1] + c >= 0 for (a, b, c) in edges]
    return torch.stack(inside).all(dim=0)


def quad_loss(pred_quads: Tensor, targets: Tensor, reduction: str = "mean") -> Tensor:
    """https://arxiv.org/abs/2103.11636"""
    batch_size = pred_quads.shape[0]
    permutations = []
    for shift in range(4):
        shifted_quad = torch.roll(targets, shifts=-shift, dims=1)
        permutations.append(shifted_quad)
        permutations.append(shifted_quad.flip(dims=(1,)))
    permutations = torch.stack(permutations, dim=1)
    targets = rearrange(permutations, "b c h w -> (b c) h w", b=batch_size, c=8)
    pred_quads = pred_quads.repeat_interleave(8, dim=0)
    loss = functional.smooth_l1_loss(pred_quads, targets, reduction="none").mean((1, 2))
    loss = reduce(rearrange(loss, "(b c) -> b c", b=batch_size, c=8), "b c -> b", "min")
    return 0.1 * loss.mean()


def uncross(quadrilaterals: Tensor) -> Tensor:
    """Take arbitrary quadrilaterals and reorder their vertices according to their angle
    to the topmost vertex.

    Args:
        quadrilaterals (Tensor[N, 4, 2]): A batch of quadrilaterals.

    Returns:
        Tensor[N, 4, 2]: The same batch of quadrilaterals, with vertices re-ordered.
    """
    n = quadrilaterals.shape[0]
    topmost_idxs = torch.argmin(quadrilaterals[:, :, 1], dim=1)
    topmost_points = quadrilaterals[torch.arange(n), topmost_idxs].unsqueeze(1)
    shifted = quadrilaterals - topmost_points
    angles = torch.atan2(shifted[:, :, 1], shifted[:, :, 0])
    angles = torch.where(angles < 0, angles + 2 * torch.pi, angles)  # wrap angles
    # set the angle of the topmost vertex to -1 to ensure it's first after sorting
    angles[torch.arange(n), topmost_idxs] = -1  # FIXME: do not mutate
    _, sorted_indices = torch.sort(angles, dim=1)
    reordered = torch.gather(
        quadrilaterals, 1, sorted_indices.unsqueeze(-1).expand(-1, -1, 2)
    )
    return reordered


def convexify(quadrilaterals: Tensor) -> Tensor:
    """Take arbitrary quadrilaterals and convexify them.

    Args:
        quadrilaterals (Tensor[N, 4, 2]): A batch of quadrilaterals.

    Returns:
        Tensor[N, 4, 2]: The same batch of quadrilaterals, with vertices re-ordered
        and adjusted for concave quadrilaterals.
    """
    uncrossed = uncross(quadrilaterals)
    edges = torch.roll(uncrossed, -1, dims=1) - uncrossed
    cross_products = (
        edges[:, :, 0] * torch.roll(edges, -1, dims=1)[:, :, 1]
        - edges[:, :, 1] * torch.roll(edges, -1, dims=1)[:, :, 0]
    )
    # For concave quads, replace the point opposite (idx 2) to the pivot (idx 0)
    # with the center of the segment bounded by the two other points (idxs 1 & 3)
    is_concave = torch.any(cross_products < 0, dim=1).reshape(-1, 1, 1)
    center_points = is_concave * ((uncrossed[:, 1] + uncrossed[:, 3]) / 2).unsqueeze(1)
    zeros = torch.zeros_like(center_points)
    center_points = torch.cat([zeros, zeros, center_points, zeros], dim=1)
    # replace c with c': [a,b,c',d] = [a,b,c,d] * [1,1,0,1] + [0,0,c',0]
    uncrossed = uncrossed * (~center_points.to(torch.bool)) + center_points
    return uncrossed



from typing import Type, Protocol, Literal, Optional

from torch import nn, Tensor


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        bias: bool = False,
        groups: int = 1,
    ):
        super().__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise = nn.Conv2d(
            in_channels, out_channels, kernel_size=1, stride=1, groups=groups, bias=bias
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.pointwise(self.depthwise(x))


class ConvNormAct(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
        padding: Optional[int] = None,
        norm: Literal["batch", "group", None] = "batch",
        act: Literal["relu", "silu", "sigmoid", "softplus", "softmax", None] = "relu",
        bias: Optional[int] = None,
        conv_layer: nn.Module = nn.Conv2d,
    ) -> None:
        conv_layer = nn.Conv2d if kernel_size == 1 else conv_layer
        layers = [
            conv_layer(
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding or ((kernel_size - 1) // 2 * dilation),
                dilation=dilation,
                groups=groups,
                bias=bias or (norm is None),
            )
        ]

        if norm == "batch":
            layers.append(nn.BatchNorm2d(out_channels))
        elif norm == "group":
            layers.append(nn.GroupNorm(32, out_channels))

        if act == "relu":
            layers.append(nn.ReLU())
        elif act == "silu":
            layers.append(nn.SiLU())
        elif act == "sigmoid":
            layers.append(nn.Sigmoid())
        elif act == "softplus":
            layers.append(nn.Softplus())
        elif act == "softmax":
            layers.append(nn.Softmax(dim=1))

        super().__init__(*layers)


class HasInOutChannels(Protocol):
    def __init__(
        self, in_channels: int, out_channels: int, *args, **kwargs
    ) -> None: ...


class SequentialConvBlocks(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_layers: int,
        kernel_size: int = 3,
        ConvBlock: Type[HasInOutChannels] = ConvNormAct,
        **kwargs,
    ) -> None:
        if num_layers <= 0:
            super().__init__(nn.Identity())
        else:
            super().__init__(
                ConvBlock(in_channels, out_channels, kernel_size=kernel_size, **kwargs),
                *[
                    ConvBlock(
                        out_channels, out_channels, kernel_size=kernel_size, **kwargs
                    )
                    for _ in range(num_layers - 1)
                ],
            )



from typing import List, Optional

from torch import Tensor, nn


class SihlModel(nn.Module):
    """A SihlModel consists of a backbone, zero or one neck, and one or more head(s)."""

    def __init__(
        self, backbone: nn.Module, neck: Optional[nn.Module], heads: List[nn.Module]
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.neck = neck
        self.heads = nn.ModuleList(heads)

    def extract_features(self, input: Tensor) -> List[Tensor]:
        x = self.backbone(input)
        if self.neck is not None:
            return self.neck(x)
        return x

    def forward(self, input: Tensor) -> List[Tensor]:
        x = self.extract_features(input)
        return [head(x) for head in self.heads]



from pathlib import Path
from typing import List, Tuple

from torch import Tensor
import numpy as np
import onnx
import onnxruntime
import pytest
import torch

from sihl.heads import QuadrilateralDetection


BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH = 4, 256, 128, 128
NUM_CLASSES, MIN_LEVEL, MAX_LEVEL = 100, 3, 7
ONNX_VERSION, ONNX_FILE_NAME = 18, "quadrilateral_detection.onnx"


@pytest.fixture()
def model() -> QuadrilateralDetection:
    return QuadrilateralDetection(
        in_channels=(3,) + (NUM_CHANNELS,) * MAX_LEVEL, num_classes=NUM_CLASSES
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [torch.randn((BATCH_SIZE, 3, HEIGHT, WIDTH))] + [
        torch.randn((BATCH_SIZE, NUM_CHANNELS, HEIGHT // 2**_, WIDTH // 2**_))
        for _ in range(MAX_LEVEL)
    ]


@pytest.fixture()
def targets() -> Tuple[List[Tensor], List[Tensor]]:
    NUM_OBJECTS = 10
    return (
        [torch.randint(0, NUM_CLASSES, (NUM_OBJECTS,)) for _ in range(BATCH_SIZE)],
        [
            torch.randint(0, HEIGHT, (NUM_OBJECTS, 4, 2), dtype=torch.float)
            for _ in range(BATCH_SIZE)
        ],
    )


def test_forward(model: QuadrilateralDetection, backbone_output: List[Tensor]) -> None:
    num_instances, scores, classes, quads = model(backbone_output)
    assert tuple(num_instances.shape) == (BATCH_SIZE,)
    assert tuple(scores.shape) == (BATCH_SIZE, model.max_instances)
    assert tuple(classes.shape) == (BATCH_SIZE, model.max_instances)
    assert tuple(quads.shape) == (BATCH_SIZE, model.max_instances, 4, 2)


def test_training_step(
    model: QuadrilateralDetection,
    backbone_output: List[Tensor],
    targets: Tuple[List[Tensor], List[Tensor]],
) -> None:
    loss, _ = model.training_step(backbone_output, *targets)
    assert loss.item()


def test_validation_step(
    model: QuadrilateralDetection,
    backbone_output: List[Tensor],
    targets: Tuple[List[Tensor], List[Tensor]],
) -> None:
    model.on_validation_start()
    loss, _ = model.validation_step(backbone_output, *targets)
    assert loss.item()
    metrics = model.on_validation_end()
    assert metrics


@pytest.fixture()
def onnx_model(model: QuadrilateralDetection, backbone_output: List[Tensor]) -> None:
    torch.onnx.export(
        model.to(torch.float32),
        args=backbone_output,
        f=ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
        output_names=[f"head0/{name}" for name, shape in model.output_shapes.items()],
        dynamic_axes=dict(
            {
                f"input_level_{lvl}": {
                    0: "batch_size",
                    2: f"height/{2**lvl}",
                    3: f"width/{2**lvl}",
                }
                for lvl in range(len(backbone_output))
            },
            **{
                f"head0/{name}": {
                    shape_idx: str(shape_value)
                    for shape_idx, shape_value in enumerate(shape)
                }
                for name, shape in model.output_shapes.items()
            },
        ),
        external_data=False,
        verify=True,
        # dynamo=True,
        # report=True,
    )
    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(
    onnx_model, model: QuadrilateralDetection, backbone_output: List[Tensor]
) -> None:
    model.eval()
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    # just check that 99% of values are equal.
    pytorch_output = [_.detach().numpy() for _ in model(backbone_output)]
    onnx_output = onnx_session.run(None, onnx_input)
    for i in range(len(pytorch_output)):
        assert (
            np.sum(np.abs(pytorch_output[i] - onnx_output[i]) > 1) / onnx_output[i].size
            < 0.01
        )



from typing import List

from torch import Tensor, nn
from torch.nn import functional
import torch


class Normalize(nn.Module):
    def __init__(self, mean: List[float], std: List[float]) -> None:
        super().__init__()
        self.register_buffer("mean", torch.tensor(mean).reshape(1, -1, 1, 1))
        self.register_buffer("std", torch.tensor(std).reshape(1, -1, 1, 1))

    def forward(self, x: Tensor) -> Tensor:
        return (x - self.mean.to(x.device)) / self.std.to(x.device)


class PadToMultipleOf(nn.Module):
    def __init__(self, n: int) -> None:
        super().__init__()
        self.n = n

    def forward(self, x: Tensor) -> Tensor:
        pad_x = (self.n - x.shape[3] % self.n) % self.n
        pad_y = (self.n - x.shape[2] % self.n) % self.n
        return functional.pad(
            x, (pad_x // 2, pad_x - pad_x // 2, pad_y // 2, pad_y - pad_y // 2)
        )



from typing import List

from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import AutoregressiveTextRecognition

from .common import get_images, plot_to_numpy


@get_images.register(AutoregressiveTextRecognition)
def _(head, config, input, target, features) -> List[np.ndarray]:
    prediction = head(features)
    if "tokens" in config:
        tokens = config["tokens"]
    else:
        tokens = [str(_) for _ in range(head.num_tokens)]

    if isinstance(tokens, str):
        tokens = list(tokens)

    tokens = tokens + [""]

    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []

    for batch, image in enumerate(images):
        blank_image = np.full_like(image, fill_value=255)

        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)

        axes[1].title.set_text("Target")
        axes[1].imshow(blank_image)
        if target is not None:
            patch_text = " ".join(
                [tokens[idx] if tokens else f"<{idx}>" for idx in target[batch]]
            )
            set_text(axes[1], patch_text.strip() + "\n\n")
        axes[2].title.set_text("Prediction")
        axes[2].imshow(blank_image)

        pred_tokens = prediction[batch]  # [:num_chars]
        patch_text = " ".join([tokens[idx] for idx in pred_tokens]).strip()
        set_text(axes[2], patch_text)

        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations


def set_text(ax, text):
    ax.text(
        0.5,
        0.5,
        text,
        transform=ax.transAxes,
        ha="center",
        va="center",
        fontsize="large",
        wrap=True,
    )


def set_text_on_peak(ax, image, text, fontsize=8):
    peak = np.unravel_index(np.argmax(image), image.shape)
    ax.text(
        peak[1],
        peak[0],
        text,
        fontsize=fontsize,
        color="black",
        ha="center",
        va="center",
        bbox={"facecolor": "white", "edgecolor": "none", "alpha": 0.5},
    )



from pathlib import Path
from random import shuffle
from typing import Tuple
import logging

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
from torchvision.io.image import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import AnomalyDetection


class MVTecDataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, train: bool, subtask: str = "cable") -> None:
        self.train = train
        self.data_dir = data_dir / subtask
        self.images = list(
            (self.data_dir / ("train" if train else "test")).rglob("*.png")
        )
        shuffle(self.images)
        self.transforms = transforms.Compose(
            [
                transforms.Resize((256, 256)),
                transforms.ToDtype(torch.float32, scale=True),
            ]
        )

    def __getitem__(self, idx: int) -> Tensor:
        image_path = self.images[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), mode=ImageReadMode.RGB
        )
        if not self.train:
            gt_dir = self.data_dir / "ground_truth" / image_path.parent.name
            gt_file = gt_dir / f"{image_path.stem}_mask.png"
            if gt_file.exists():
                mask = torchvision.io.read_image(str(gt_file.resolve())).to(torch.bool)
                image, mask = self.transforms(image, tv_tensors.Mask(mask))
                return image, mask.squeeze(0)
        image = self.transforms(image)
        mask = torch.zeros(image.shape[1:], dtype=torch.bool)
        return image, mask

    def __len__(self) -> int:
        return len(self.images)


class MVTecDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int, subtask: str = "leather") -> None:
        super().__init__()
        self.batch_size = batch_size
        self.subtask = subtask
        self.data_dir = Path(__file__).parent / "data" / "MVTec"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "ipythonx/mvtec-ad", path=self.data_dir, unzip=True, quiet=False
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = MVTecDataset(self.data_dir, train=True, subtask=self.subtask)
        self.valset = MVTecDataset(self.data_dir, train=False, subtask=self.subtask)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="anomaly_detection"
    )
    lightning.seed_everything(0)

    datamodule = MVTecDataModule(batch_size=1, subtask="cable")
    datamodule.prepare_data()
    datamodule.setup()
    trainer = pl.Trainer(
        max_steps=70_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
        log_every_n_steps=len(datamodule.trainset) // datamodule.batch_size,
    )

    with trainer.init_module():
        LEVEL = 3
        backbone = TorchvisionBackbone(
            "resnet50",
            pretrained=True,
            top_level=LEVEL,
            frozen_levels=-1,
            freeze_batchnorms=True,
        )
        head = AnomalyDetection(in_channels=backbone.out_channels, level=LEVEL)
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
        )

    if hasattr(head, "pretraining_step"):
        head.eval()
        with torch.no_grad():
            model.on_pretraining_start()
            for batch in datamodule.train_dataloader():
                model.pretraining_step(batch)
            model.on_pretraining_end()
        head.train()

    log.debug(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=datamodule)



from typing import Union, Optional, Tuple
from torch import Tensor, nn

from einops import reduce
from sihl.layers.convblocks import ConvNormAct
from sihl.layers.pooling import BlurPool2d
from sihl.utils import interpolate


class StridedDownscaler(ConvNormAct):
    def __init__(self, in_channels: int, out_channels: int, **kwargs) -> None:
        super().__init__(in_channels, out_channels, stride=2, **kwargs)


class AntialiasedDownscaler(nn.Sequential):
    def __init__(
        self, in_channels: int, out_channels: int, kernel_size: int = 3, **kwargs
    ) -> None:
        super().__init__(
            ConvNormAct(in_channels, out_channels, kernel_size, **kwargs),
            BlurPool2d(out_channels, stride=2),
        )


class SimpleDownscaler(nn.Sequential):
    def __init__(
        self, in_channels: int, out_channels: int, kernel_size: int = 3, **kwargs
    ) -> None:
        super().__init__(
            ConvNormAct(in_channels, out_channels, kernel_size, **kwargs),
            nn.AvgPool2d(2, stride=2),
        )


class BilinearScaler(nn.Module):
    def __init__(
        self,
        scale: Optional[Union[float, int]] = None,
        size: Optional[Union[int, Tuple[int, int]]] = None,
    ) -> None:
        super().__init__()
        assert scale or size, "Specify scale or size."
        self.scale = scale
        self.size = size

    def forward(self, x: Tensor) -> Tensor:
        if self.scale is not None and float(self.scale) == 1.0:
            return x
        return interpolate(x, scale_factor=self.scale, size=self.size)


class SimpleUpscaler(nn.Sequential):
    def __init__(
        self, in_channels: int, out_channels: int, kernel_size: int = 3
    ) -> None:
        super().__init__(
            BilinearScaler(scale=2), ConvNormAct(in_channels, out_channels, kernel_size)
        )


class BilinearAdditiveUpscaler(nn.Module):
    """[The Devil is in the Decoder](https://arxiv.org/abs/1707.05847)"""

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3):
        super().__init__()
        self.upscaler = BilinearScaler(scale=2)
        self.residual = nn.ConvTranspose2d(
            in_channels, in_channels // 4, kernel_size=2, stride=2
        )
        self.out_conv = ConvNormAct(
            in_channels // 4, out_channels, kernel_size=kernel_size
        )

    def forward(self, x: Tensor) -> Tensor:
        a = reduce(self.upscaler(x), "b (c1 c) h w -> b c h w", "mean", c1=4)
        b = self.residual(x)
        return self.out_conv(a + b)



from typing import List, Tuple, Dict
import math

from torch import nn, Tensor
from torch.nn import functional
from torchmetrics import MeanMetric
from torchmetrics.retrieval import RetrievalPrecision, RetrievalRPrecision
import torch

from sihl.utils import EPS


class MetricLearning(nn.Module):
    """Metric learning is mapping semantically similar images to close points in
    embedding space. It is trained like a multiclass classification head, with pairs of
    sample and corresponding "identities" (equivalent to classes). However, at inference
    time, the metric learning head produces embedding vectors instead of identity
    predictions. These embeddings can then be compared to one another using the eucliean
    distance, enabling clustering, k-NN classification, or other downstream tasks.

    Refs:
        1. [Arcface](https://arxiv.org/abs/1801.07698)

    Todo:
        - [Center Contrastive Loss](https://arxiv.org/abs/2308.00458)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_identities: int,
        embedding_dim: int = 256,
        level: int = 5,
        margin: float = 0.5,
        num_subcenters: int = 1,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_identities (int): Number of possible semantic identities.
            embedding_dim (int, optional): Number of embedding channels. Defaults to 256.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
            margin (float, optional): Margin hyperparameter (c.f. [1]). Defaults to 0.5.
            num_subcenters (int, optional): Number of clusters per identity (c.f. [1]). Defaults to 1.
        """
        assert num_identities > 0, num_identities
        assert len(in_channels) > level, (len(in_channels), level)
        assert embedding_dim > 0, embedding_dim
        super().__init__()

        self.num_identities = num_identities
        self.level = level
        self.num_subcenters = num_subcenters
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels[level], embedding_dim, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )
        self.scale = math.sqrt(2) * math.log(num_identities - 1)
        self.margin = margin
        self.weight = nn.Parameter(
            torch.FloatTensor(num_subcenters, embedding_dim, num_identities)
        )
        nn.init.xavier_uniform_(self.weight)
        self.index_embeddings, self.index_ids = None, None

        self.output_shapes = {"embeddings": ("batch_size", embedding_dim)}

    def forward(self, inputs: List[Tensor]) -> Tensor:
        return functional.normalize(self.conv(inputs[self.level]))

    def training_step(self, inputs: List[Tensor], targets: Tensor) -> Tensor:
        feats = self.forward(inputs)
        feats = feats.unsqueeze(0).expand(self.num_subcenters, *feats.shape)
        cos_theta = torch.max(
            torch.bmm(feats, functional.normalize(self.weight)), dim=0
        ).values
        theta = torch.acos(torch.clamp(cos_theta, -1 + EPS, 1 - EPS))
        one_hot = torch.zeros_like(cos_theta)
        one_hot.scatter_(1, targets.view(-1, 1).long(), 1)
        selected = torch.where(
            theta > math.pi - self.margin, torch.zeros_like(one_hot), one_hot
        ).bool()
        logits = torch.where(selected, theta + self.margin, theta).cos() * self.scale
        return functional.cross_entropy(logits, targets), {}

    def reset_validation_index_set(self) -> None:
        self.index_embeddings = None
        self.index_ids = None

    def extend_validation_index_set(self, inputs: List[Tensor], ids: Tensor) -> None:
        if self.index_embeddings is None:
            self.index_embeddings = self.forward(inputs)
            self.index_ids = ids
        else:
            self.index_embeddings = torch.cat(
                [self.index_embeddings, self.forward(inputs)]
            )
            self.index_ids = torch.cat([self.index_ids, ids])

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        # https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Precision
        self.precision_computers = {}
        for k in [1, 3, 5]:
            self.precision_computers[k] = RetrievalPrecision(top_k=k)
        self.r_precision = RetrievalRPrecision()
        self.sample_counter = 0
        self.knn_matches = {}
        for k in [1, 3, 5]:
            self.knn_matches[k] = 0

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        embeddings = self.forward(inputs)
        batch_size, device = embeddings.shape[0], embeddings.device
        similarities = functional.linear(embeddings, weight=self.index_embeddings)
        top5_values, top5_targets = torch.topk(
            similarities, k=min(similarities.shape[1], 6)
        )
        top5_targets = self.index_ids[top5_targets]
        indexes = torch.arange(batch_size, device=device).unsqueeze(1)
        preds = top5_values[:, 1:]  # 1:6 instead of 0:5 to remove self similarity
        target = top5_targets[:, 1:] == targets.unsqueeze(1)
        indexes = self.sample_counter + indexes.repeat(1, preds.shape[1])
        for k in [1, 3, 5]:
            self.precision_computers[k].to(device).update(preds, target, indexes)
        self.r_precision.to(device).update(preds, target, indexes)
        self.sample_counter += batch_size
        for k in [1, 3, 5]:
            self.knn_matches[k] += target[:, :k].sum()
        # loss, _ = self.training_step(inputs, targets)  # loss increases ?!
        self.loss_computer.to(device).update(0)
        return torch.zeros((1,)), {}

    def on_validation_end(self) -> Dict[str, float]:
        metrics = {"loss": self.loss_computer.compute().item()}
        for k in [1, 3, 5]:
            metrics[f"precision_at_{k}"] = self.precision_computers[k].compute().item()
        metrics["r_precision"] = self.r_precision.compute().item()
        for k in [1, 3, 5]:
            metrics[f"{k}nn_accuracy"] = self.knn_matches[k] / self.sample_counter / k
        return metrics



from typing import Final, List, Dict, Type

from torch import Tensor, nn
from torchvision.models.feature_extraction import create_feature_extractor
import torch
import torchvision

from sihl.layers import AntialiasedDownscaler, Normalize
from sihl.utils import recursive_setattr, recursive_getattr, interpolate


TORCHVISION_MODEL_LEVELS: Final[Dict[str, List[str]]] = {
    "convnext_base": [f"features.{idx}" for idx in [0, 1, 3, 5, 7]],
    "convnext_large": [f"features.{idx}" for idx in [0, 1, 3, 5, 7]],
    "convnext_small": [f"features.{idx}" for idx in [0, 1, 3, 5, 7]],
    "convnext_tiny": [f"features.{idx}" for idx in [0, 1, 3, 5, 7]],
    "densenet121": ["features.relu0"]
    + [f"features.denseblock{_}" for _ in [1, 2, 3, 4]],
    "densenet161": ["features.relu0"]
    + [f"features.denseblock{_}" for _ in [1, 2, 3, 4]],
    "densenet169": ["features.relu0"]
    + [f"features.denseblock{_}" for _ in [1, 2, 3, 4]],
    "efficientnet_b0": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b1": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b2": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b3": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b4": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b5": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b6": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_b7": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_v2_l": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_v2_m": [f"features.{_}" for _ in [1, 2, 3, 5, 8]],
    "efficientnet_v2_s": [f"features.{_}" for _ in [1, 2, 3, 5, 7]],
    "mnasnet0_5": [f"layers.{_}" for _ in [7, 8, 9, 11, 16]],
    "mnasnet0_75": [f"layers.{_}" for _ in [7, 8, 9, 11, 16]],
    "mnasnet1_0": [f"layers.{_}" for _ in [7, 8, 9, 11, 16]],
    "mnasnet1_3": [f"layers.{_}" for _ in [7, 8, 9, 11, 16]],
    "mobilenet_v2": [f"features.{_}" for _ in [1, 3, 6, 13, 18]],
    "mobilenet_v3_large": [f"features.{_}" for _ in [1, 3, 6, 12, 16]],
    "mobilenet_v3_small": [f"features.{_}" for _ in [0, 1, 3, 8, 12]],
    "resnet101": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnet152": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnet18": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnet34": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnet50": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnext101_32x8d": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnext101_64x4d": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "resnext50_32x4d": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "shufflenet_v2_x0_5": ["conv1", "maxpool", "stage2", "stage3", "conv5"],
    "shufflenet_v2_x1_0": ["conv1", "maxpool", "stage2", "stage3", "conv5"],
    "shufflenet_v2_x1_5": ["conv1", "maxpool", "stage2", "stage3", "conv5"],
    "shufflenet_v2_x2_0": ["conv1", "maxpool", "stage2", "stage3", "conv5"],
    "wide_resnet50_2": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
    "wide_resnet101_2": ["relu"] + [f"layer{_}" for _ in [1, 2, 3, 4]],
}
FIRST_LAYERS: Final[Dict[str, str]] = {
    "convnext_base": "features.0.0",
    "convnext_large": "features.0.0",
    "convnext_small": "features.0.0",
    "convnext_tiny": "features.0.0",
    "densenet121": "features.conv0",
    "densenet161": "features.conv0",
    "densenet169": "features.conv0",
    "efficientnet_b0": "features.0.0",
    "efficientnet_b1": "features.0.0",
    "efficientnet_b2": "features.0.0",
    "efficientnet_b3": "features.0.0",
    "efficientnet_b4": "features.0.0",
    "efficientnet_b5": "features.0.0",
    "efficientnet_b6": "features.0.0",
    "efficientnet_b7": "features.0.0",
    "efficientnet_v2_l": "features.0.0",
    "efficientnet_v2_m": "features.0.0",
    "efficientnet_v2_s": "features.0.0",
    "mnasnet0_5": "layers.0",
    "mnasnet0_75": "layers.0",
    "mnasnet1_0": "layers.0",
    "mnasnet1_3": "layers.0",
    "mobilenet_v2": "features.0.0",
    "mobilenet_v3_large": "features.0.0",
    "mobilenet_v3_small": "features.0.0",
    "resnet101": "conv1",
    "resnet152": "conv1",
    "resnet18": "conv1",
    "resnet34": "conv1",
    "resnet50": "conv1",
    "resnext101_32x8d": "conv1",
    "resnext101_64x4d": "conv1",
    "resnext50_32x4d": "conv1",
    "shufflenet_v2_x0_5": "conv1.0",
    "shufflenet_v2_x1_0": "conv1.0",
    "shufflenet_v2_x1_5": "conv1.0",
    "shufflenet_v2_x2_0": "conv1.0",
    "wide_resnet50_2": "conv1",
    "wide_resnet101_2": "conv1",
}
TORCHVISION_BACKBONE_NAMES = tuple(TORCHVISION_MODEL_LEVELS.keys())


class TorchvisionBackbone(nn.Module):
    def __init__(
        self,
        name: str,
        pretrained: bool = False,
        input_channels: int = 3,
        top_level: int = 5,
        frozen_levels: int = 0,
        freeze_batchnorms: bool = False,
        downscaler: Type[nn.Module] = AntialiasedDownscaler,
    ) -> None:
        """
        Args:
            name (str): Architecture name (must be in TORCHVISION_BACKBONE_NAMES)
            pretrained (bool, optional): Use weights from Imagenet pretraining. Defaults to False.
            input_channels (int, optional): Number of input image channels. Defaults to 3.
            top_level (int, optional): Deepest level (with stride 2^level). Defaults to 5.
            frozen_levels (int, optional): How many levels to freeze (from level 0). Defaults to 0.
            freeze_batchnorms (bool, optional): Whether to freeze batchnorm running stats. Defaults to False.
            downscaler (Type[nn.Module], optional): Downscaler module used for levels above 5. Defaults to AntialiasedDownscaler.
        """
        super().__init__()
        self.name = name
        self.top_level = top_level
        try:
            level_names = TORCHVISION_MODEL_LEVELS[name][: self.top_level]
        except KeyError as error:
            raise ValueError(
                f"Architecture {name} is not supported. "
                f"Select from {TORCHVISION_BACKBONE_NAMES}"
            ) from error
        self.normalize = nn.Identity()
        if pretrained and input_channels == 3:
            self.normalize = Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )
        self.model = torchvision.models.get_model(
            name, weights=("DEFAULT" if pretrained else None)
        )
        self.model = create_feature_extractor(self.model, level_names)

        first_conv = recursive_getattr(self.model, FIRST_LAYERS[name])
        if input_channels != first_conv.in_channels:
            recursive_setattr(
                self.model,
                FIRST_LAYERS[name],
                nn.Conv2d(
                    in_channels=input_channels,
                    out_channels=first_conv.out_channels,
                    kernel_size=first_conv.kernel_size,
                    stride=first_conv.stride,
                    padding=first_conv.padding,
                    dilation=first_conv.dilation,
                    groups=first_conv.groups,
                    bias=first_conv.bias,
                    padding_mode=first_conv.padding_mode,
                ),
            )

        # freeze modules in first `frozen_levels` levels
        if pretrained:
            freeze_levels(self.model, frozen_levels, level_names, freeze_batchnorms)

        min_size = 2 ** (self.top_level + 1)
        self.dummy_input = torch.zeros(1, input_channels, min_size, min_size)
        self.out_channels = [input_channels] + [
            _.shape[1] for _ in self.model(self.dummy_input).values()
        ]
        out_channels = self.out_channels[-1]
        extra_layers = range(self.top_level - 5)
        self.out_channels += [out_channels for _ in extra_layers]
        self.downscalers = nn.ModuleList(
            [AntialiasedDownscaler(out_channels, out_channels) for _ in extra_layers]
        )

    def forward(self, input: Tensor) -> List[Tensor]:
        assert input.shape[2] % 2**self.top_level == 0
        assert input.shape[3] % 2**self.top_level == 0
        x = self.normalize(input)
        outputs = list(self.model(x).values())
        outputs = [input] + [
            interpolate(output, size=(x.shape[2] // 2**level, x.shape[3] // 2**level))
            for output, level in zip(outputs, range(1, self.top_level + 1))
        ]
        for downscaler in self.downscalers:
            outputs.append(downscaler(outputs[-1]))
        return outputs


def freeze_levels(
    model: nn.Module,
    num_levels: int,
    level_names: List[str],
    freeze_batchnorms: bool = False,
) -> None:
    """freeze modules in first `num_levels` levels"""
    if num_levels < 0:
        for param in model.parameters():
            param.requires_grad_(False)
    elif num_levels > 0:
        num_levels = min(num_levels, len(level_names))
        last_level_to_freeze = level_names[num_levels - 1]
        last_level_reached = False
        freezing = True
        for module_name, _ in model.named_modules():
            if module_name == "":
                continue
            if module_name == last_level_to_freeze:
                last_level_reached = True
            if last_level_reached and last_level_to_freeze not in module_name:
                freezing = False
            for param in model.get_submodule(module_name).parameters():
                param.requires_grad_(not freezing)
    else:
        for param in model.parameters():
            param.requires_grad_(True)

    if freeze_batchnorms:
        for layer in model.modules():
            if isinstance(layer, (nn.BatchNorm2d, nn.LayerNorm, nn.GroupNorm)):
                layer.eval()



from pathlib import Path
import logging

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import InstanceSegmentation
from sihl.layers import FPN

coco_labels = [
    "__background__",
    "person",
    "bicycle",
    "car",
    "motorbike",
    "aeroplane",
    "bus",
    "train",
    "truck",
    "boat",
    "trafficlight",
    "firehydrant",
    "streetsign",
    "stopsign",
    "parkingmeter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "hat",
    "backpack",
    "umbrella",
    "shoe",
    "eyeglasses",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sportsball",
    "kite",
    "baseballbat",
    "baseballglove",
    "skateboard",
    "surfboard",
    "tennisracket",
    "bottle",
    "plate",
    "wineglass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hotdog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "sofa",
    "pottedplant",
    "bed",
    "mirror",
    "diningtable",
    "window",
    "desk",
    "toilet",
    "door",
    "tvmonitor",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cellphone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "blender",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddybear",
    "hairdrier",
    "toothbrush",
    "hairbrush",
]

valid_coco_labels = coco_labels.copy()
for _ in [
    "__background__",
    "streetsign",
    "hat",
    "shoe",
    "eyeglasses",
    "plate",
    "mirror",
    "window",
    "desk",
    "door",
    "blender",
    "hairbrush",
]:
    valid_coco_labels.remove(_)


def collate_fn(batch):
    classes, masks = [], []
    for sample in batch:
        annot = sample[1]
        if "labels" in annot and "masks" in annot:
            sample_classes, sample_masks = [], []
            for label, mask in zip(annot["labels"], annot["masks"]):
                if mask.max() > 0:  # ensure no mask is empty
                    sample_classes.append(valid_coco_labels.index(coco_labels[label]))
                    sample_masks.append(mask.to(torch.float32))
            classes.append(torch.tensor(sample_classes))
            masks.append(torch.stack(sample_masks))
        else:
            classes.append(torch.tensor([]))
            masks.append(torch.tensor([]))
    return torch.stack([_[0] for _ in batch]), {"classes": classes, "masks": masks}


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context"
            )

    def setup(self, stage: str = "") -> None:
        annot_dir = self.data_dir / "annotations_trainval2017"
        self.trainset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "train2017"),
                str(annot_dir / "instances_train2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            ),
            target_keys=["masks", "labels"],
        )
        self.valset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "val2017"),
                str(annot_dir / "instances_val2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            ),
            target_keys=["masks", "labels"],
        )

    def train_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="instance_segmentation"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=90_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        precision="16-mixed",
        val_check_interval=1000,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True, frozen_levels=1)
        neck = FPN(backbone.out_channels, out_channels=256, bottom_level=3, top_level=7)
        head = InstanceSegmentation(
            in_channels=neck.out_channels,
            num_classes=len(valid_coco_labels),
            mask_top_level=5,
            soft_label_decay_steps=90_000,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.SGD,
            optimizer_kwargs={"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
            scheduler=torch.optim.lr_scheduler.MultiStepLR,
            scheduler_kwargs={
                "milestones": [60_000, 80_000],
                "gamma": 0.1,
                "warmup_batches": 1000,
            },
            data_config={"categories": valid_coco_labels},
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))



from pathlib import Path
from typing import Tuple, List, Literal, Dict, Any
import logging
import random
import sys
import xml.etree.ElementTree as ET

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

sys.path.append("../src")

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import QuadrilateralDetection


AIRCRAFT_LABELS = [f"A{n}" for n in range(1, 21)]


### HACK: torchvision doesn't support augmenting keypoints yet
# https://pytorch.org/blog/extending-torchvisions-transforms-to-object-detection-segmentation-and-video-tasks/#development-milestones-and-future-work
# So, we'll convert each keypoint to a tiny square bbox
# we'll augment those then convert them back by taking their center point
def polygons_to_bboxes(polygons: Tensor, canvas_size: Tuple[int, int]):
    """polygons: (N, K, 2)"""
    box_centers = polygons.reshape((-1, 2))
    flat_boxes = torch.cat([box_centers, torch.ones_like(box_centers)], dim=1)
    return tv_tensors.BoundingBoxes(
        flat_boxes, format=tv_tensors.BoundingBoxFormat.CXCYWH, canvas_size=canvas_size
    )


def bboxes_to_polygons(bboxes: tv_tensors.BoundingBoxes, num_vertices: int) -> Tensor:
    return bboxes[:, :2].reshape((-1, num_vertices, 2))


###


def collate_fn(batch: List[Tuple[Tensor, Dict[Literal["classes", "boxes"], Any]]]):
    return (
        torch.stack([sample[0] for sample in batch]),
        {
            "classes": [
                (torch.tensor([AIRCRAFT_LABELS.index(_) for _ in sample[1]["classes"]]))
                for sample in batch
            ],
            "quads": [sample[1]["quads"] for sample in batch],
        },
    )


def parse_xml_annot(raw: str) -> Tuple[List[str], Tensor]:
    root = ET.fromstring(raw)
    categories = []
    quads = []
    for obj in root.findall("object"):
        categories.append(obj.find("name").text)
        quad = obj.find("robndbox")
        x1 = float(quad.find("x_left_top").text)
        y1 = float(quad.find("y_left_top").text)
        x2 = float(quad.find("x_right_top").text)
        y2 = float(quad.find("y_right_top").text)
        x3 = float(quad.find("x_right_bottom").text)
        y3 = float(quad.find("y_right_bottom").text)
        x4 = float(quad.find("x_left_bottom").text)
        y4 = float(quad.find("y_left_bottom").text)
        quads.append([[x1, y1], [x2, y2], [x3, y3], [x4, y4]])
    return categories, torch.tensor(quads)


class AircraftDataset(torch.utils.data.Dataset[Tuple[Tensor, bool]]):
    def __init__(self, data_dir: Path, train: bool, indices: List[str]) -> None:
        self.train = train
        self.images: List[Tuple[Path, Tuple[str, Tensor]]] = []
        images_dir = data_dir / "JPEGImages"
        annots_dir = data_dir / "Annotations" / "Oriented Bounding Boxes"
        # split = "train" if self.train else "test"
        # with open(data_dir / "ImageSets" / "Main" / f"{split}.txt") as f:
        #     idxs = f.read().splitlines()
        for idx in indices:
            with open(annots_dir / f"{idx}.xml") as f:
                categories, quads = parse_xml_annot(f.read())
            self.images.append(
                (images_dir / f"{idx}.jpg", {"classes": categories, "quads": quads})
            )
        self.transforms = None
        # if self.train:
        # https://github.com/pytorch/vision/issues/566#issuecomment-814221986
        if self.train:
            self.transforms = transforms.Compose(
                [
                    transforms.Resize(512 - 1, max_size=512),
                    transforms.RandomCrop(512, pad_if_needed=True),
                    transforms.RandomHorizontalFlip(p=0.5),
                    transforms.RandomVerticalFlip(p=0.5),
                    transforms.RandomChoice(
                        [
                            transforms.RandomRotation((90, 90)),
                            transforms.RandomRotation((180, 180)),
                            transforms.RandomRotation((270, 270)),
                            transforms.Identity(),
                        ]
                    ),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transforms = transforms.Compose(
                [
                    transforms.Resize(512 - 1, max_size=512),
                    transforms.RandomCrop(512, pad_if_needed=True),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target = self.images[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), torchvision.io.ImageReadMode.RGB
        )
        quad_boxes = polygons_to_bboxes(target["quads"], image.shape[1:])
        image, quad_boxes = self.transforms(image, quad_boxes)
        target["quads"] = bboxes_to_polygons(quad_boxes, num_vertices=4)
        return image, target

    def __len__(self) -> int:
        return len(self.images)


class AircraftDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "military_aircraft_recognition"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "khlaifiabilel/military-aircraft-recognition-dataset",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        # Re-split to 80/20
        all_idxs = [_.stem for _ in (self.data_dir / "JPEGImages").glob("*.jpg")]
        random.shuffle(all_idxs)
        split_idxs = int(0.2 * len(all_idxs))
        train_idxs, valid_idxs = all_idxs[split_idxs:], all_idxs[:split_idxs]
        self.trainset = AircraftDataset(self.data_dir, train=True, indices=train_idxs)
        self.valset = AircraftDataset(self.data_dir, train=False, indices=valid_idxs)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, bool]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            collate_fn=collate_fn,
            drop_last=True,
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, bool]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="quadrilateral_detection"
    )
    lightning.seed_everything(0)

    MAX_STEPS = 5_000
    trainer = pl.Trainer(
        max_steps=MAX_STEPS,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = QuadrilateralDetection(
            backbone.out_channels,
            num_classes=len(AIRCRAFT_LABELS),
            soft_label_decay_steps=MAX_STEPS,
            bottom_level=5,
            top_level=5,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.SGD,
            optimizer_kwargs={"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
            scheduler=torch.optim.lr_scheduler.MultiStepLR,
            scheduler_kwargs={"milestones": [MAX_STEPS * 3 // 4], "gamma": 0.1},
            data_config={"categories": AIRCRAFT_LABELS},
        )

    log.debug(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=AircraftDataModule(batch_size=64))



from typing import Any, Tuple, List, Dict

from einops import rearrange, reduce
from torch import Tensor
from torch.nn import functional
import torch

from sihl.utils import EPS

from .instance_segmentation import InstanceSegmentation


class KeypointDetection(InstanceSegmentation):
    """Keypoint detection is like object detection, except instances are all of the same
    category (i.e. `num_classes` = 1), and are associated with a set of 2D (key)points
    instead of an axis-aligned bounding box. Each instance predictions must provide a 2D
    coordinate for every keypoint, as well as whether each keypoint is present in the
    image or not (i.e. missing or invisible). Coordinate predictions of non-present
    keypoints are safe to ignore.

    Refs:
        1. [FCPose](https://arxiv.org/abs/2105.14185)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_keypoints: int,
        bottom_level: int = 3,
        top_level: int = 7,
        num_channels: int = 256,
        num_layers: int = 4,
        max_instances: int = 32,
        t_min: float = 0.2,
        t_max: float = 0.6,
        topk: int = 7,
        soft_label_decay_steps: int = 1,
        mask_top_level: int = 5,
        mask_num_channels: int = 32,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_keypoints (int): Number of keypoints.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            max_instances (int, optional): Maximum number of instances to predict in a sample. Defaults to 32.
            t_min (float, optional): Lower bound of O2F parameter t. Defaults to 0.2.
            t_max (float, optional): Upper bound of O2F parameter t. Defaults to 0.6.
            topk (int, optional): How many anchors to match to each ground truth object when copmuting the loss. Defaults to 7.
            soft_label_decay_steps (int, optional): How many training steps to perform before the one-to-few matching becomes one-to-one. Defaults to 1.
            mask_top_level (int, optional): Top level of inputs masks are computed from. Defaults to 5.
            mask_num_channels (int, optional): Number of convolutional channels in the keypoint heatmap prediction branch.
        """
        super().__init__(
            num_classes=1,
            in_channels=in_channels,
            num_channels=num_channels,
            num_layers=num_layers,
            bottom_level=bottom_level,
            top_level=top_level,
            max_instances=max_instances,
            t_min=t_min,
            t_max=t_max,
            topk=topk,
            soft_label_decay_steps=soft_label_decay_steps,
            mask_top_level=mask_top_level,
            mask_num_channels=mask_num_channels,
            mask_out_channels=num_keypoints,
        )
        self.num_kpts = num_keypoints
        self.output_shapes = {
            "num_instances": ("batch_size",),
            "instance_scores": ("batch_size", max_instances),
            "keypoint_scores": ("batch_size", max_instances, num_keypoints),
            "keypoints": ("batch_size", max_instances, num_keypoints, 2),
        }

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        num_instances, scores, _, logits = self.forward_logits(inputs)
        b, _, h, w = logits.shape
        heatmaps = rearrange(
            logits, "b (i k) h w -> b i k h w", i=self.max_instances, k=self.num_kpts
        )
        heatmaps = heatmaps.flatten(3, 4).softmax(dim=3).reshape(heatmaps.shape)
        keypoint_scores = reduce(heatmaps, "b i k h w -> b i k", "max")
        keypoints = self.heatmap_to_keypoints(inputs, heatmaps)
        return num_instances, scores, keypoint_scores, keypoints

    def heatmap_to_keypoints(self, inputs: List[Tensor], heatmaps: Tensor) -> Tensor:
        heatmap_height, heatmap_width = heatmaps.shape[3:]
        kpt_pos = rearrange(heatmaps, "b i k h w -> b i k (h w)").argmax(dim=3)
        img_height, img_width = inputs[0].shape[2:]
        y = ((kpt_pos // heatmap_width) / heatmap_height) * img_height
        x = ((kpt_pos % heatmap_width) / heatmap_width) * img_width
        return torch.stack([x, y], dim=-1)

    def training_step(
        self,
        inputs: List[Tensor],
        keypoints: List[Tensor],
        presence: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, Any]]:
        batch_size, _, img_height, img_width = inputs[0].shape
        scores, reg_features = self.get_features(inputs)
        controllers = self.get_controllers(reg_features)
        pred_boxes = self.get_boxes(reg_features)
        bottom_features = self.get_bottom_features(inputs)
        anchors = self.get_anchors(inputs)  # (2, H, W)
        normalized_anchors = self.get_anchors(inputs, normalized=True)

        boxes = [keypoints_to_boxes(k, p) for (k, p) in zip(keypoints, presence)]
        classes = [
            torch.zeros((k.shape[0],), dtype=torch.int64, device=scores.device)
            for k in keypoints
        ]

        class_target, box_target, assignment = self.get_targets(
            anchors, scores, pred_boxes, classes, boxes, is_validating
        )
        class_loss, box_loss = self.get_losses(
            scores, pred_boxes, class_target, box_target
        )

        heatmap_losses, distance_losses = [], []
        heatmap_height, heatmap_width = bottom_features.shape[2:]
        for batch_idx in range(batch_size):
            pos_idxs = class_target[batch_idx].max(dim=1).values >= 1 - EPS
            if keypoints[batch_idx].numel() == 0 or not torch.any(pos_idxs):
                continue

            anchor_idxs = assignment[batch_idx, pos_idxs]
            target_coords = keypoints[batch_idx][anchor_idxs].clone()
            target_coords[:, :, 0] *= (heatmap_width - 1) / (img_width - 1)
            target_coords[:, :, 0].clamp_(0, heatmap_width - 1)
            target_coords[:, :, 1] *= (heatmap_height - 1) / (img_height - 1)
            target_coords[:, :, 1].clamp_(0, heatmap_height - 1)
            target_coords = target_coords.round().to(torch.int64)
            gt_xs = functional.one_hot(target_coords[:, :, 0], heatmap_width)
            gt_ys = functional.one_hot(target_coords[:, :, 1], heatmap_height)

            target_heatmap = (gt_xs.unsqueeze(2) * gt_ys.unsqueeze(3)).to(class_loss)
            pred_heatmap = self.get_masks(
                params=controllers[batch_idx][pos_idxs],
                anchors=normalized_anchors[pos_idxs],
                bottom_features=bottom_features[batch_idx],
                num_masks=pos_idxs.sum(),
            ).reshape((-1, self.num_kpts, heatmap_height, heatmap_width))

            presence_mask = presence[batch_idx][anchor_idxs].flatten()
            shape = (-1, heatmap_height * heatmap_width)  # CE over flattened space
            flat_pred_hm = pred_heatmap.reshape(shape)[presence_mask]
            flat_target_hm = target_heatmap.reshape(shape)[presence_mask]
            with torch.autocast(device_type="cuda", enabled=False):
                ce_loss = functional.cross_entropy(
                    flat_pred_hm.to(torch.float32), flat_target_hm, reduction="none"
                )
                heatmap_losses.append(ce_loss.nan_to_num(0).mean())

            peaks = flat_pred_hm.argmax(1)
            peak_rows, peaks_cols = peaks // heatmap_width, peaks % heatmap_width
            pred_keypoints = torch.stack(
                [peaks_cols / (heatmap_width - 1), peak_rows / (heatmap_height - 1)],
                dim=1,
            )
            gt_keypoints = target_coords.flatten(0, 1)[presence_mask].to(pred_keypoints)
            gt_keypoints[:, 0] /= heatmap_width - 1
            gt_keypoints[:, 1] /= heatmap_height - 1
            distance_losses.append((gt_keypoints - pred_keypoints).pow(2).mean())

        heatmap_loss = torch.stack(heatmap_losses).mean() if heatmap_losses else 0
        distance_loss = torch.stack(distance_losses).mean() if distance_losses else 0

        return class_loss + box_loss + heatmap_loss + distance_loss, {
            "class_loss": class_loss,
            "box_loss": box_loss,
            "heatmap_loss": heatmap_loss,
            "distance_loss": distance_loss,
        }

    def validation_step(
        self,
        inputs: List[Tensor],
        keypoints: List[Tensor],
        presence: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, Any]]:
        boxes = [keypoints_to_boxes(*args) for args in zip(keypoints, presence)]
        _, scores, kpt_scores, pred_keypoints = self.forward(inputs)
        pred_boxes = [
            keypoints_to_boxes(kpts, pres > 0.1)
            for (kpts, pres) in zip(pred_keypoints, kpt_scores)
        ]
        dev = scores.device
        self.map_computer.to(dev).update(
            [
                {"scores": s, "labels": torch.zeros_like(s, dtype=int), "boxes": b}
                for s, b in zip(scores, pred_boxes)
            ],
            [
                {"labels": torch.zeros(b.shape[0], dtype=int, device=dev), "boxes": b}
                for b in boxes
            ],
        )
        loss, metrics = self.training_step(inputs, keypoints, presence, True)
        self.loss_computer.to(dev).update(loss)
        return loss, metrics


def keypoints_to_boxes(keypoints: Tensor, presence: Tensor) -> Tensor:
    assert presence.dtype == torch.bool
    masked_keypoints = keypoints.clone()
    masked_keypoints[~presence] = torch.inf
    xmin = masked_keypoints[:, :, 0].min(dim=1).values
    ymin = masked_keypoints[:, :, 1].min(dim=1).values
    masked_keypoints[~presence] = -torch.inf
    xmax = masked_keypoints[:, :, 0].max(dim=1).values
    ymax = masked_keypoints[:, :, 1].max(dim=1).values
    return torch.stack([xmin, ymin, xmax, ymax], dim=-1)


# FIXME
# def compute_oks(
#     pred_keypoints: Tensor, gt_keypoints: Tensor, presence: Tensor
# ) -> Tensor:
#     dists = torch.sum((pred_keypoints - gt_keypoints) ** 2, dim=2)  # Shape: (B, K)
#     bboxes = keypoints_to_boxes(gt_keypoints, presence)
#     areas = ((bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])).unsqueeze(1)
#     sigmas = torch.full((gt_keypoints.shape[1],), 0.05, device=pred_keypoints.device)
#     exp_terms = torch.exp(-dists / (sigmas**2) * 2 * (areas**2)) * presence  # (B, K)
#     oks_scores = exp_terms.sum(dim=1) / presence.sum(dim=1)
#     return oks_scores



# ruff: noqa: F401
from typing import Any, Tuple, List, Callable, Union, Optional, Literal
import functools
import random
import sys

from torch import nn, Tensor
from torch.nn import functional
from torch.nn.functional import avg_pool2d
import torch
import torchvision

from sihl.utils.polygon_iou import polygon_iou


EPS = 1e-5
interpolate = functools.partial(functional.interpolate, mode="bilinear")


class BatchedMeanVarianceAccumulator:
    def __init__(self):
        self.count = 0
        self.mean = None
        self.M2 = None

    def update(self, x: Tensor) -> None:
        if self.mean is None:
            self.mean = x.mean(dim=0)
            self.M2 = torch.zeros_like(self.mean)
        else:
            if x.shape[1:] != self.mean.shape:
                raise ValueError(
                    f"Shape mismatch: got {x.shape[1:]}, expected {self.mean.shape}"
                )

            batch_count = x.size(0)
            total_count = self.count + batch_count
            batch_mean = x.mean(dim=0)
            batch_delta = batch_mean - self.mean
            self.mean += batch_delta * batch_count / total_count
            # Welford's method
            self.M2 += (
                x.var(dim=0, unbiased=False) * batch_count
                + (batch_delta**2) * self.count * batch_count / total_count
            )

        self.count += x.size(0)

    def compute(self):
        """
        Compute the mean and variance based on the accumulated data.
        """
        if self.count < 2:
            return self.mean, torch.full_like(self.mean, float("nan"))
        variance = self.M2 / (self.count - 1)
        return self.mean, variance


# def init(m: nn.Module, kaiming: bool = True) -> None:
#     if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
#         if kaiming:
#             nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
#         else:
#             nn.init.xavier_normal_(m.weight)
#         if m.bias is not None:
#             nn.init.constant_(m.bias, 0)

#     elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
#         nn.init.constant_(m.weight, 1)
#         nn.init.constant_(m.bias, 0)


def random_pad(
    image: Tensor, target_size: Union[int, Tuple[int, int]], fill: Union[float, int] = 0
) -> Tensor:
    image_size = image.shape[1:]
    if isinstance(target_size, int):
        target_size = (target_size, target_size)

    if image_size[0] > target_size[0] or image_size[1] > target_size[1]:
        scale_factor = min(
            target_size[0] / image_size[0], target_size[1] / image_size[1]
        )
        new_height = int(image_size[0] * scale_factor)
        new_width = int(image_size[1] * scale_factor)
        image = torchvision.transforms.functional.resize(image, (new_height, new_width))
        image_size = image.shape[1:]

    pad_width = target_size[1] - image_size[1]
    pad_height = target_size[0] - image_size[0]

    top_pad = random.randint(0, pad_height)
    bottom_pad = pad_height - top_pad
    left_pad = random.randint(0, pad_width)
    right_pad = pad_width - left_pad

    padded_image = torchvision.transforms.functional.pad(
        image, (left_pad, top_pad, right_pad, bottom_pad), fill
    )
    return padded_image


def coordinate_grid(height: int, width: int, y_max: float, x_max: float) -> Tensor:
    """2D grid of pixel center coordinates (0 to x_max, 0 to y_max)."""
    y_max, x_max = torch.scalar_tensor(y_max), torch.scalar_tensor(x_max)
    y_min, x_min = y_max / height / 2, x_max / width / 2
    ys = torch.linspace(y_min, y_max - y_min, steps=height)
    xs = torch.linspace(x_min, x_max - x_min, steps=width)
    return torch.stack([xs[None, :].repeat(height, 1), ys[:, None].repeat(1, width)])


def f_score(beta: float) -> Callable[[Tensor, Tensor], Tensor]:
    """https://en.wikipedia.org/wiki/F-score#Definition"""
    return lambda p, r: (1 + beta**2) * p * r / (beta**2 * p + r)


def points_to_bbox(points: Tensor) -> Tensor:
    """(N, K, 2) -> (N, 4)"""
    min_x = torch.min(points[..., 0], dim=1).values
    min_y = torch.min(points[..., 1], dim=1).values
    max_x = torch.max(points[..., 0], dim=1).values
    max_y = torch.max(points[..., 1], dim=1).values
    return torch.stack([min_x, min_y, max_x, max_y], dim=1)


def edges(x: Tensor) -> Tensor:
    sobel_kernel_x = torch.tensor(
        [[[[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]]], dtype=x.dtype, device=x.device
    )
    sobel_kernel_y = torch.tensor(
        [[[[-1, -2, -1], [0, 0, 0], [1, 2, 1]]]], dtype=x.dtype, device=x.device
    )
    sobel_kernel_x = sobel_kernel_x.repeat(x.shape[1], 1, 1, 1)
    sobel_kernel_y = sobel_kernel_y.repeat(x.shape[1], 1, 1, 1)
    edges_x = functional.conv2d(x, sobel_kernel_x, padding=1, groups=x.shape[1])
    edges_y = functional.conv2d(x, sobel_kernel_y, padding=1, groups=x.shape[1])
    edges = torch.sqrt(edges_x**2 + edges_y**2)
    edges = edges / edges.max()
    return edges


def gaussian_blur(
    x: Tensor, kernel_size: int = 5, sigma: Optional[float] = None
) -> Tensor:
    sigma = sigma or 0.3 * ((kernel_size - 1) * 0.5 - 1) + 0.8
    kernel_1d = torch.exp(
        -(torch.arange(-(kernel_size // 2), kernel_size // 2 + 1) ** 2) / (2 * sigma**2)
    )
    kernel_1d = kernel_1d / kernel_1d.sum()
    kernel_2d = torch.outer(kernel_1d, kernel_1d).view(1, 1, kernel_size, kernel_size)
    kernel_2d = kernel_2d.expand(x.shape[1], -1, -1, -1).to(x)
    return functional.conv2d(x, kernel_2d, padding=kernel_size // 2, groups=x.shape[1])


def ssim_loss(
    pred: Tensor, gt: Tensor, window_size: int = 11, size_average: bool = True
) -> Tensor:
    C1, C2 = 0.01**2, 0.03**2
    padding = window_size // 2
    mu1 = avg_pool2d(pred, window_size, stride=1, padding=padding)
    mu2 = avg_pool2d(gt, window_size, stride=1, padding=padding)
    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2
    sigma1_sq = avg_pool2d(pred * pred, window_size, stride=1, padding=padding) - mu1_sq
    sigma2_sq = avg_pool2d(gt * gt, window_size, stride=1, padding=padding) - mu2_sq
    sigma12 = avg_pool2d(pred * gt, window_size, stride=1, padding=padding) - mu1_mu2
    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / (
        (mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2)
    )
    return (1 - ssim_map.mean() if size_average else 1 - ssim_map.sum()).abs()


def focal_loss(
    preds: Tensor, targets: Tensor, alpha: float = 0.25, gamma: float = 2.0
) -> Tensor:
    """https://pytorch.org/vision/main/_modules/torchvision/ops/focal_loss.html"""
    ce_loss = functional.binary_cross_entropy(preds, targets, reduction="none")
    p_t = preds * targets + (1 - preds) * (1 - targets)
    loss = ce_loss * ((1 - p_t) ** gamma)
    alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
    return alpha_t * loss


def tversky_loss(
    preds: Tensor,
    targets: Tensor,
    alpha: float = 0.5,
    beta: float = 0.5,
    ignore_index: int = -100,
) -> torch.Tensor:
    valid_mask = targets != ignore_index
    targets = functional.one_hot(targets * valid_mask, preds.shape[1])
    targets = targets.permute(0, 3, 1, 2) * valid_mask.unsqueeze(1)
    preds = functional.softmax(preds, dim=1) * valid_mask.unsqueeze(1)
    tp = (preds * targets).sum(dim=(2, 3))
    fn = ((1 - preds) * targets).sum(dim=(2, 3))
    fp = (preds * (1 - targets)).sum(dim=(2, 3))
    return 1 - ((tp + EPS) / (tp + alpha * fp + beta * fn + EPS)).mean()


def recursive_getattr(obj: Any, attr: str, *args):
    """https://stackoverflow.com/a/31174427"""

    def _getattr(obj: Any, attr: str):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def recursive_setattr(obj: Any, attr: str, val: Any):
    """https://stackoverflow.com/a/31174427"""
    pre, _, post = attr.rpartition(".")
    return setattr(recursive_getattr(obj, pre) if pre else obj, post, val)


def call_and_get_locals(func: Callable, *args, **kwargs) -> Tuple[Any, List[Any]]:
    """https://stackoverflow.com/a/52358426"""
    frame = None

    def snatch_locals(_frame, name, arg):
        nonlocal frame
        if name == "call":
            frame = _frame
            sys.settrace(sys._getframe(0).f_trace)
        return sys._getframe(0).f_trace

    sys.settrace(snatch_locals)

    try:
        result = func(*args, **kwargs)
    finally:
        sys.settrace(sys._getframe(0).f_trace)

    return result, frame.f_locals



# ruff: noqa: F401
try:
    from importlib_metadata import version
except ImportError:
    from importlib.metadata import version


from sihl.sihl_model import SihlModel
from sihl.torchvision_backbone import TorchvisionBackbone, TORCHVISION_BACKBONE_NAMES

__version__ = version("sihl")

try:
    from sihl.timm_backbone import TimmBackbone, TIMM_BACKBONE_NAMES
except ImportError as e:
    print(e)
    pass

try:
    from sihl.lightning_module import SihlLightningModule
except ImportError as e:
    print(e)
    pass



# ruff: noqa: F401, F811
from typing import Any

from torch import Tensor

from sihl.sihl_model import SihlModel
from sihl.heads import ViewInvarianceLearning

from .common import get_images
from . import anomaly_detection
from . import autoencoding
from . import autoregressive_text_recognition
from . import depth_estimation
from . import instance_segmentation
from . import keypoint_detection
from . import metric_learning
from . import multiclass_classification
from . import multilabel_classification
from . import object_detection
from . import panoptic_segmentation
from . import quadrilateral_detection
from . import regression
from . import scene_text_recognition
from . import semantic_segmentation
from . import view_invariance_learning


def visualize(
    model: SihlModel,
    configs: Any,
    input: Tensor,
    targets: Any,
    logger: Any,
    step: int,
    start_idx: int = 0,
):
    features = model.eval().extract_features(input)
    for idx, (config, head, target) in enumerate(zip(configs, model.heads, targets)):
        if isinstance(head, ViewInvarianceLearning):  # FIXME: this is a hack
            target = model.extract_features(target[: features[0].shape[0]])
        vizs = get_images(head, config, input, target, features)  # singly dispatched
        for viz_idx, viz_img in enumerate(vizs):
            logger.experiment.add_image(
                f"{idx}/visualizations/{start_idx + viz_idx}", viz_img, global_step=step
            )



import math
from typing import Optional, List, Tuple, Dict

from einops import rearrange
from torch import nn, Tensor
from torch.nn import functional
from torch.nn.modules.transformer import _generate_square_subsequent_mask
from torchmetrics.text import WordErrorRate
import torch


class AutoregressiveTextRecognition(nn.Module):
    """
    Autoregressive text recognition is like scene text recognition, but with a
    sequential, autoregressive approach. This means that the latency is higher, but the
    accuracy tends to be better, especially for longer sequences.

    Refs:
        1. (Holistic Representation Guided Attention Network)[https://arxiv.org/abs/1904.01375]
    """

    def __init__(
        self,
        in_channels: List[int],
        num_tokens: int,
        max_sequence_length: int,
        level: int = 3,
        num_channels: int = 256,
        num_layers: int = 1,
        num_heads: int = 4,
        embedding_dim: int = 1024,
        dropout: float = 0.0,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_tokens (int): Number of possible text tokens.
            max_sequence_length (int): Maximum number of tokens to predict in a single sample.
            level (int, optional): Level of inputs this head is attached to. Defaults to 3.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 1.
            num_heads (int, optional): Number of attention heads. Defaults to 4.
            embedding_dim (int, optional): Embedding dimensionality for tokens. Defaults to 1024.
            dropout (float, optional): Dropout rate. Defaults to 0.0.
        """
        assert num_tokens > 0
        assert max_sequence_length > 0
        assert level < len(in_channels)
        super().__init__()

        self.in_channels = in_channels
        self.num_tokens = num_tokens
        self.max_sequence_length = max_sequence_length
        self.level = level
        self.pad, self.begin, self.end = num_tokens, num_tokens + 1, num_tokens + 2

        self.visual_encoding = nn.Sequential(
            nn.Conv2d(in_channels[level], num_channels, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )
        self.lateral_conv = nn.Conv2d(in_channels[level], num_channels, kernel_size=1)
        self.embedding = nn.Embedding(self.num_tokens + 3, num_channels, self.pad)
        self.positional_encoding = PositionalEncoding(num_channels, dropout=dropout)
        self.decoder = CausalTransformerDecoder(
            CausalTransformerDecoderLayer(
                d_model=num_channels,
                nhead=num_heads,
                dim_feedforward=embedding_dim,
                dropout=dropout,
                norm_first=True,
            ),
            num_layers=num_layers,
        )
        self.fc = nn.Linear(num_channels, self.num_tokens + 3, bias=False)
        # TODO: cloze refinement and/or bidirectional decoding
        self.output_shapes = {"tokens": ("batch_size", max_sequence_length)}

    def get_next_token(
        self,
        sequence: Tensor,
        visual_encoding: Tensor,
        memory: Tensor,
        cache: Optional[Tensor],
    ) -> Tensor:
        sequence = sequence.to(torch.int64)
        y = self.positional_encoding(self.embedding(sequence) + visual_encoding)
        output, cache = self.decoder(y, memory, cache)
        output = self.fc(output)
        pred_tokens = torch.argmax(output, dim=2)[-1, :]
        pred_tokens[sequence[-1, :] == self.end] = self.end
        sequence = torch.cat([sequence, pred_tokens.unsqueeze(0)])
        return sequence, visual_encoding, memory, cache

    def forward(self, inputs: List[Tensor]) -> Tensor:
        x = inputs[self.level]
        visual_encoding = self.visual_encoding(x).unsqueeze(0)
        batch_size, device = x.shape[0], visual_encoding.device
        memory = rearrange(self.lateral_conv(x), "b c h w -> (h w) b c")
        sequence = torch.full((1, batch_size), self.begin, device=device)
        cache = None
        for idx in range(1, self.max_sequence_length + 1):
            sequence, visual_encoding, memory, cache = self.get_next_token(
                sequence, visual_encoding, memory, cache
            )
            # if torch.all(sequence[-1, :] == self.end):  # FIXME: ONNX export
            #     break
        sequence[sequence == self.begin] = self.pad
        sequence[sequence == self.end] = self.pad
        sequence = rearrange(sequence[1:], "l b -> b l")  # [1:] removes [BEGIN]
        return sequence

    def training_step(
        self, inputs: List[Tensor], texts: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        x = inputs[self.level]
        visual_encoding = self.visual_encoding(x).unsqueeze(0)
        memory = rearrange(self.lateral_conv(x), "b c h w -> (h w) b c")

        target = pad_tensors(texts, pad=self.end).to(torch.int64).to(memory.device)
        target = functional.pad(target, (0, 1), value=self.end)  # (N, S)
        sequence = functional.pad(target[:, :-1], (1, 0), value=self.begin)  # (N, S)

        y = self.positional_encoding(self.embedding(sequence.t()) + visual_encoding)
        logits = self.fc(self.decoder.training_step(y, memory)).permute(1, 2, 0)
        loss = functional.cross_entropy(logits, target, reduction="mean")
        return loss, {}

    @torch.no_grad()
    def on_validation_start(self) -> None:
        self.token_error_rate = WordErrorRate()
        self.matches: List[bool] = []

    @torch.no_grad()
    def validation_step(
        self, inputs: List[Tensor], texts: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        pred_texts = self.forward(inputs)
        predictions = [
            " ".join(str(token.item()) for token in text if token != self.pad)
            for text in pred_texts
        ]
        ground_truths = [
            " ".join(str(token.item()) for token in text) for text in texts
        ]
        self.token_error_rate.update(predictions, ground_truths)
        self.matches.extend(
            [pred == gt for pred, gt in zip(predictions, ground_truths)]
        )
        return self.training_step(inputs, texts)

    @torch.no_grad()
    def on_validation_end(self) -> Dict[str, float]:
        return {
            "token_error_rate": self.token_error_rate.compute().item(),
            "accuracy": sum(self.matches) / max(len(self.matches), 1),
        }


class PositionalEncoding(nn.Module):
    """https://pytorch.org/tutorials/beginner/transformer_tutorial.html"""

    def __init__(self, d_model: int, dropout: float = 0, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        """x shape: (seq_len, batch_size, embedding_dim)"""
        return self.dropout(x + self.pe[: x.shape[0]])


class CausalTransformerDecoder(nn.TransformerDecoder):
    """https://github.com/alex-matton/causal-transformer-decoder"""

    def forward(
        self,
        tgt: Tensor,
        memory: Optional[Tensor] = None,
        cache: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """tgt shape: (seq_len x batch_size, embedding_dim)"""
        output = tgt

        new_token_cache = []
        for i, mod in enumerate(self.layers):
            output = mod(output, memory)
            new_token_cache.append(output)
            if cache is not None:
                output = torch.cat([cache[i], output])

        if cache is not None:
            new_cache = torch.cat([cache, torch.stack(new_token_cache)], dim=1)
        else:
            new_cache = torch.stack(new_token_cache)

        return output, new_cache

    def training_step(self, tgt: Tensor, memory: Optional[Tensor] = None) -> Tensor:
        output = tgt
        for mod in self.layers:
            output = mod.training_step(output, memory)
        return output


class CausalTransformerDecoderLayer(nn.TransformerDecoderLayer):
    def forward(self, tgt: Tensor, memory: Optional[Tensor] = None) -> Tensor:
        tgt_last_tok = tgt[-1:, :, :]
        tgt_last_tok = self.norm1(tgt_last_tok)  # NOTE: pre-LN
        tmp_tgt = self.self_attn(tgt_last_tok, tgt, tgt, attn_mask=None)[0]
        tgt_last_tok = tgt_last_tok + self.dropout1(tmp_tgt)

        if memory is not None:
            tgt_last_tok = self.norm2(tgt_last_tok)
            # TODO: get attention maps for visualization
            tmp_tgt = self.multihead_attn(tgt_last_tok, memory, memory)[0]
            tgt_last_tok = tgt_last_tok + self.dropout2(tmp_tgt)

        tgt_last_tok = self.norm3(tgt_last_tok)
        tmp_tgt = self.linear2(
            self.dropout(self.activation(self.linear1(tgt_last_tok)))
        )
        tgt_last_tok = tgt_last_tok + self.dropout3(tmp_tgt)
        return tgt_last_tok

    def training_step(self, tgt: Tensor, memory: Optional[Tensor] = None) -> Tensor:
        return super().forward(
            tgt,
            memory,
            tgt_mask=_generate_square_subsequent_mask(tgt.size(0), tgt.device),
        )


def pad_tensors(tensors: List[Tensor], pad: int) -> Tensor:
    assert len(tensors)
    max_length = max(len(t) for t in tensors)
    padded = [
        functional.pad(t.unsqueeze(0), (0, max_length - len(t)), value=pad).squeeze(0)
        for t in tensors
    ]
    return torch.stack(padded)



from typing import List
from pathlib import Path

from torch import Tensor
import onnx
import onnxruntime
import pytest
import torch

from sihl.layers import FPN


BATCH_SIZE, TOP_CHANNELS, TOP_HEIGHT, TOP_WIDTH = 4, 1024, 16, 16
OUT_CHANNELS = 64
NUM_CLASSES, NUM_OBJECTS = 10, 5
BOTTOM_LEVEL, TOP_LEVEL = 3, 7
ONNX_FILE_NAME, ONNX_VERSION = "fpn.onnx", 18


@pytest.fixture()
def model() -> FPN:
    return FPN(
        in_channels=[3] + [TOP_CHANNELS // 2**_ for _ in range(5)][::-1],
        out_channels=OUT_CHANNELS,
        bottom_level=BOTTOM_LEVEL,
        top_level=TOP_LEVEL,
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [
        torch.randn(BATCH_SIZE, 3, TOP_HEIGHT * 32, TOP_WIDTH * 32),
        torch.randn(BATCH_SIZE, TOP_CHANNELS // 16, TOP_HEIGHT * 16, TOP_WIDTH * 16),
        torch.randn(BATCH_SIZE, TOP_CHANNELS // 8, TOP_HEIGHT * 8, TOP_WIDTH * 8),
        torch.randn(BATCH_SIZE, TOP_CHANNELS // 4, TOP_HEIGHT * 4, TOP_WIDTH * 4),
        torch.randn(BATCH_SIZE, TOP_CHANNELS // 2, TOP_HEIGHT * 2, TOP_WIDTH * 2),
        torch.randn(BATCH_SIZE, TOP_CHANNELS, TOP_HEIGHT, TOP_WIDTH),
    ]


def test_forward(model: FPN, backbone_output: List[Tensor]) -> None:
    outputs = model(backbone_output)
    assert len(outputs) == TOP_LEVEL + 1
    for output in outputs[BOTTOM_LEVEL : TOP_LEVEL + 1]:
        assert output.shape[1] == OUT_CHANNELS


def test_levels(backbone_output: List[Tensor]) -> None:
    model = FPN(
        in_channels=[3] + [TOP_CHANNELS // 2**_ for _ in range(5)][::-1],
        out_channels=OUT_CHANNELS,
        bottom_level=len(backbone_output) - 3,
        top_level=len(backbone_output) - 2,
    )
    outputs = model(backbone_output)
    assert len(outputs) == len(backbone_output)
    assert outputs[len(backbone_output) - 1].shape[1] == backbone_output[-1].shape[1]
    assert outputs[len(backbone_output) - 2].shape[1] == OUT_CHANNELS
    assert outputs[len(backbone_output) - 3].shape[1] == OUT_CHANNELS


@pytest.fixture()
def onnx_model(model: FPN, backbone_output: List[Tensor]) -> None:
    torch.onnx.export(
        model,
        backbone_output,
        ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
    )

    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(onnx_model, model: FPN, backbone_output: List[Tensor]) -> None:
    model.eval()
    with torch.no_grad():
        pytorch_output = model(backbone_output)
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    onnx_output = onnx_session.run(None, onnx_input)
    for idx in range(len(onnx_output)):
        torch.testing.assert_close(pytorch_output[idx].numpy(), onnx_output[idx])



from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
from torchvision.transforms.v2 import Resize
import numpy as np
import torch

from sihl.heads import PanopticSegmentation

from .common import get_images, plot_to_numpy, COLORS


@get_images.register(PanopticSegmentation)
def _(head, config, input, target, features) -> List[np.ndarray]:
    categories = config["categories"] if "categories" in config else None
    prediction = head.forward(features)[:, 0, :, :]  # semantic map
    target = target[:, 0, :, :]  # semantic map
    resize = Resize(target[0].shape[1:], antialias=False)
    images = (resize(input).permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()

    visualizations = []
    for batch, image in enumerate(images):
        seen_categories = []
        fig, axes = plt.subplots(1, 3, figsize=(10, 5), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)

        def get_map_img(pixelmap):
            shape = (pixelmap.shape[0], pixelmap.shape[1], 4)
            map_img = np.full(shape, fill_value=255, dtype=np.uint8)
            for idx in pixelmap.unique():
                if idx == config.get("ignore", None):
                    continue
                label = str(idx) if categories is None else categories[idx]
                if label not in seen_categories:
                    seen_categories.append(label)
                color = COLORS[seen_categories.index(label) % len(COLORS)]
                mask = (pixelmap == idx).to(torch.uint8)
                map_img[..., 0] = mask * color[0] + (1 - mask) * map_img[..., 0]
                map_img[..., 1] = mask * color[1] + (1 - mask) * map_img[..., 1]
                map_img[..., 2] = mask * color[2] + (1 - mask) * map_img[..., 2]
                map_img[..., 3] = mask * 255 + (1 - mask) * map_img[..., 3]
            return map_img

        axes[1].title.set_text("Target")
        if target is not None:
            axes[1].imshow(get_map_img(target[batch].to("cpu")))
        axes[2].title.set_text("Prediction")
        if prediction is not None:
            axes[2].imshow(get_map_img(prediction[batch].to("cpu")))

        fig.legend(
            handles=[
                patches.Patch(
                    color=[_ / 255 for _ in COLORS[i % len(COLORS)]], label=label
                )
                for i, label in enumerate(seen_categories)
            ],
            loc="lower center",
            frameon=False,
            ncol=min(7, len(seen_categories)),
        )
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from functools import partial
from typing import List, Dict, Tuple, Optional

from einops import rearrange, reduce
from torch import nn, Tensor
from torchmetrics import MeanMetric, JaccardIndex, Accuracy
import torch

from sihl.layers import (
    SimpleDownscaler,
    SimpleUpscaler,
    ConvNormAct,
    BilinearScaler,
    SequentialConvBlocks,
)
from sihl.utils import interpolate, BatchedMeanVarianceAccumulator

sequential_downscalers = partial(SequentialConvBlocks, ConvBlock=SimpleDownscaler)
sequential_upscalers = partial(SequentialConvBlocks, ConvBlock=SimpleUpscaler)


class AnomalyDetection(nn.Module):
    """Anomaly detection is telling whether an image contains an "anomaly" or not. The
    difference with binary classification is that this head only needs "positive samples"
    (i.e. "normal" images) to train on, whereas the binary classification head would
    need those and a similar amount of "negative samples" (i.e. "anomalous" images).
    This task is self-supervised, but it needs labeled samples (positive and negative)
    for validation.

    As described in [1], this head has 3 "submodels":
    1. a pre-trained frozen teacher, which extracts generic image features
    2. an autoencoder, which extracts compact features by attempting to recreate the
       input through a bottleneck
    3. a student model, which tries to match the outputs of the previous 2 submodels

    At prediction time, the difference between the student and the teacher highlights
    structural (local) anomalies, while the difference between the student and the
    autoencoder shows logical (global) anomalies.

    Refs:
        1. [EfficientAD](https://arxiv.org/abs/2303.14535)
    """

    def __init__(
        self,
        in_channels: List[int],
        level: int = 2,
        num_channels: int = 256,
        num_layers: int = 1,
        autoencoder_channels: int = 64,
        autoencoder_top_level: int = 5,
    ):
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            level (int, optional): Top level of inputs this head is attached to. Defaults to 3.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 3.
            autoencoder_channels (int, optional): Number of channels in the compact representation. Defaults to 64.
            autoencoder_top_level (int, optional): Top level of inputs the autoencoder is attached to. Defaults to 5.
        """
        assert num_channels > 0 and num_layers > 0
        assert len(in_channels) > level > 0
        super().__init__()

        self.level = level
        self.num_channels = num_channels
        self.ae_channels = autoencoder_channels
        self.num_layers = num_layers
        self.p_hard = 0.999
        self.autoencoder_top_level = autoencoder_top_level
        self.out_channels = in_channels[level]

        self.student = nn.Sequential(
            ConvNormAct(in_channels[0], num_channels),
            sequential_downscalers(num_channels, num_channels, num_layers=level),
            SequentialConvBlocks(num_channels, num_channels, num_layers=num_layers),
            nn.Conv2d(num_channels, self.out_channels * 2, kernel_size=3, padding=1),
        )

        self.autoencoder_encoder = nn.Sequential(
            ConvNormAct(in_channels[0], self.ae_channels),
            sequential_downscalers(
                self.ae_channels, self.ae_channels, num_layers=autoencoder_top_level
            ),
        )
        size = 8
        self.autoencoder_resize = BilinearScaler(size=(size, size))
        self.autoencoder_bottleneck = nn.Sequential(
            nn.Linear(size * size * self.ae_channels, self.ae_channels),
            nn.Linear(self.ae_channels, size * size * self.ae_channels),
        )
        upscale_levels = autoencoder_top_level - level
        self.autoencoder_decoder = nn.Sequential(
            sequential_upscalers(
                self.ae_channels, self.ae_channels, num_layers=upscale_levels
            ),
            SequentialConvBlocks(
                self.ae_channels, self.ae_channels, num_layers=num_layers
            ),
            nn.Conv2d(self.ae_channels, self.out_channels, kernel_size=3, padding=1),
        )

        self.inputs0, self.inputs_level = [], []
        self.register_buffer("local_thresh", torch.tensor([0.05]))
        self.register_buffer("global_thresh", torch.tensor([0.05]))
        self.register_buffer("features_mean", torch.tensor([0]))
        self.register_buffer("feature_std", torch.tensor([1]))
        self.register_buffer("q_st_start", torch.tensor(0))
        self.register_buffer("q_st_end", torch.tensor(0.1))
        self.register_buffer("q_ae_start", torch.tensor(0))
        self.register_buffer("q_ae_end", torch.tensor(0.1))
        self.output_shapes = {
            "anomaly_maps": ("batch_size", f"height/{2**level}", f"width/{2**level}")
        }

    def compute_distances(self, inputs: List[Tensor]) -> Tensor:
        teacher_out = (inputs[self.level] - self.features_mean) / self.feature_std
        student_out = self.student(inputs[0])

        encoded = self.autoencoder_encoder(inputs[0])
        old_size = encoded.shape[2:]
        encoded = self.autoencoder_resize(encoded)
        h, w = encoded.shape[2:]
        encoded = rearrange(encoded, "b c h w -> b (c h w)")
        encoded = rearrange(
            self.autoencoder_bottleneck(encoded), "b (c h w) -> b c h w", h=h, w=w
        )
        autoencoder_out = self.autoencoder_decoder(interpolate(encoded, size=old_size))

        distance_ae = (autoencoder_out - teacher_out).pow(2)
        distance_st = (teacher_out - student_out[:, : self.out_channels]).pow(2)
        distance_stae = (autoencoder_out - student_out[:, self.out_channels :]).pow(2)
        return distance_st, distance_ae, distance_stae

    def forward(self, inputs: List[Tensor]) -> Tensor:
        distance_st, distance_ae, distance_stae = self.compute_distances(inputs)
        local_anomaly = reduce(distance_st, "b c h w -> b 1 h w", "mean")
        local_anomaly = self.local_thresh * (
            (local_anomaly - self.q_st_start) / (self.q_st_end - self.q_st_start)
        )
        global_anomaly = reduce(distance_stae, "b c h w -> b 1 h w", "mean")
        global_anomaly = self.global_thresh * (
            (global_anomaly - self.q_ae_start) / (self.q_ae_end - self.q_ae_start)
        )
        anomaly = (local_anomaly.relu() + global_anomaly.relu()).clamp(0, 1)
        return interpolate(anomaly, size=inputs[0].shape[2:]).squeeze(1)

    def training_step(
        self,
        inputs: List[Tensor],
        targets: Optional[Tensor] = None,
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, float]]:
        if not is_validating:
            self.inputs0.append(inputs[0])
            self.inputs_level.append(inputs[self.level])
        distance_st, distance_ae, distance_stae = self.compute_distances(inputs)
        loss_st = torch.cat(
            [x[x >= torch.quantile(x, q=self.p_hard)] for x in distance_st]
        ).mean()
        loss_ae = distance_ae.mean()
        loss_stae = distance_stae.mean()
        return loss_st + loss_ae + loss_stae, {
            "loss_student_teacher": loss_st,
            "loss_autoencoder_teacher": loss_ae,
            "loss_student_autoencoder": loss_stae,
        }

    def on_validation_start(self) -> None:
        if len(self.inputs0):
            st_distances, stae_distances = [], []
            for inputs0, inputs_level in zip(self.inputs0, self.inputs_level):
                distance_st, distance_ae, distance_stae = self.compute_distances(
                    [inputs0] + [None for _ in range(self.level - 1)] + [inputs_level]
                )
                st_distances.append(reduce(distance_st, "b c h w -> b h w", "mean"))
                stae_distances.append(reduce(distance_stae, "b c h w -> b h w", "mean"))
            self.inputs0.clear()
            self.inputs_level.clear()
            # https://github.com/pytorch/pytorch/issues/64947
            distance_st = torch.cat(st_distances).flatten()[-(2**24 - 1) :]
            distance_stae = torch.cat(stae_distances).flatten()[-(2**24 - 1) :]
            self.q_st_start = torch.quantile(distance_st, q=0.9)
            self.q_st_end = torch.quantile(distance_st, q=0.995)
            self.q_ae_start = torch.quantile(distance_stae, q=0.9)
            self.q_ae_end = torch.quantile(distance_stae, q=0.995)

        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.mean_iou_computer = JaccardIndex(task="binary")
        self.accuracy = Accuracy(task="binary")

    def validation_step(
        self, inputs: List[Tensor], targets: Optional[Tensor] = None
    ) -> Tuple[Tensor, Dict[str, float]]:
        distance_st, distance_ae, distance_stae = self.compute_distances(inputs)
        loss, metrics = self.training_step(inputs, is_validating=True)
        if targets is not None:
            pred = self.forward(inputs)
            self.mean_iou_computer.to(loss.device).update(pred, targets)
            self.accuracy.to(loss.device).update(
                (pred > 0.5).any(dim=(1, 2)), targets.any(dim=(1, 2))
            )
        self.loss_computer.to(loss.device).update(loss)
        return loss, metrics

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "mean_iou": self.mean_iou_computer.compute().item(),
            "accuracy": self.accuracy.compute().item(),
        }

    def on_pretraining_start(self) -> None:
        self.feature_accumulator = BatchedMeanVarianceAccumulator()

    def pretraining_step(
        self, inputs: List[Tensor], targets: Optional[Tensor] = None
    ) -> None:
        features = rearrange(inputs[self.level], "n c h w -> (n h w) c")
        self.feature_accumulator.update(features)

    def on_pretraining_end(self):
        mean, variance = self.feature_accumulator.compute()
        self.features_mean = rearrange(mean, "c -> 1 c 1 1")
        self.feature_std = rearrange(variance.sqrt(), "c -> 1 c 1 1")



from typing import Any, Tuple, List, Dict

from einops import rearrange, repeat, reduce
from torch import nn, Tensor
from torch.nn import functional
from torchmetrics import JaccardIndex, Accuracy
from torchvision.ops import masks_to_boxes
import torch

from sihl.layers import ConvNormAct, SimpleUpscaler
from sihl.utils import interpolate

from .instance_segmentation import InstanceSegmentation


class PanopticSegmentation(InstanceSegmentation):
    """Panoptic segmentation assigns each pixel of the input to a category and an
    instance index, essentially merging instance segmentation and semantic segmentation
    tasks in one (although, in this case, instances cannot overlap).

    Refs:
        1. [CondInst](https://arxiv.org/abs/2102.03026)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_stuff_classes: int,
        num_thing_classes: int,
        bottom_level: int = 3,
        top_level: int = 7,
        num_channels: int = 256,
        num_layers: int = 4,
        max_instances: int = 100,
        t_min: float = 0.2,
        t_max: float = 0.6,
        topk: int = 7,
        soft_label_decay_steps: int = 1,
        mask_top_level: int = 5,
        ignore_index: int = -100,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_stuff_classes (int): Number of possible stuff categories.
            num_thing_classes (int): Number of possible thing categories.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            max_instances (int, optional): Maximum number of instances to predict in a sample. Defaults to 100.
            t_min (float, optional): Lower bound of O2F parameter t. Defaults to 0.2.
            t_max (float, optional): Upper bound of O2F parameter t. Defaults to 0.6.
            topk (int, optional): How many anchors to match to each ground truth object when copmuting the loss. Defaults to 7.
            soft_label_decay_steps (int, optional): How many training steps to perform before the one-to-few matching becomes one-to-one. Defaults to 1.
            mask_top_level (int, optional): Top level of inputs masks are computed from. Defaults to 5.
        """
        super().__init__(
            in_channels=in_channels,
            num_classes=num_thing_classes,
            mask_top_level=mask_top_level,
            soft_label_decay_steps=soft_label_decay_steps,
            num_channels=num_channels,
            num_layers=num_layers,
            bottom_level=bottom_level,
            top_level=top_level,
            max_instances=max_instances,
            t_min=t_min,
            t_max=t_max,
            topk=topk,
        )

        self.ignore_index = ignore_index  # https://pytorch.org/docs/stable/generated/torch.nn.CrossEntropyLoss.html
        self.num_stuff_classes = num_stuff_classes
        self.num_thing_classes = num_thing_classes
        self.semantic_branch = nn.ModuleList(
            [ConvNormAct(in_channels[bottom_level], self.num_channels)]
            + [
                nn.Sequential(
                    ConvNormAct(in_channels[level], self.num_channels),
                    *[
                        SimpleUpscaler(self.num_channels, self.num_channels)
                        for _ in range(level - bottom_level)
                    ],
                )
                for level in range(bottom_level + 1, self.mask_top_level + 1)
            ]
        )
        self.semantic_out_conv = nn.Conv2d(self.num_channels, num_stuff_classes, 1)
        self.output_shapes = {"panoptic_maps": ("batch_size", 2, "height", "width")}

    def get_semantic_map(self, inputs: List[Tensor]) -> Tensor:
        xs = inputs[self.bottom_level : self.mask_top_level + 1]
        outs = [conv(x) for x, conv in zip(xs, self.semantic_branch)]
        return self.semantic_out_conv(torch.stack(outs).sum(dim=0))

    def forward(self, inputs: List[Tensor]) -> Tensor:
        num_instances, scores, classes, masks = super().forward(inputs)
        things = self.instance_to_panoptic(scores, classes, masks)
        stuff = interpolate(
            self.get_semantic_map(inputs), size=inputs[0].shape[2:]
        ).argmax(dim=1, keepdim=True)
        stuff = stuff * (things[:, 0:1] == 0)  # zero-out thing pixels
        return things + stuff  # stuff ids are just their categories

    def panoptic_to_instance(self, maps: Tensor) -> Tuple[List[Tensor], List[Tensor]]:
        instance_classes, instance_masks = [], []
        for panoptic_map in maps:
            semantic_map, id_map = panoptic_map
            sample_instance_masks, sample_instance_classes = [], []
            for instance_id in id_map.unique():
                instance_mask = id_map == instance_id
                category_idx = semantic_map[instance_mask][0]
                if category_idx == self.ignore_index:
                    continue
                if category_idx < self.num_stuff_classes:
                    continue
                sample_instance_classes.append(category_idx - self.num_stuff_classes)
                sample_instance_masks.append(instance_mask)
            if len(sample_instance_masks):
                instance_classes.append(torch.stack(sample_instance_classes))
                instance_masks.append(torch.stack(sample_instance_masks))
            else:
                instance_classes.append(torch.empty((0,), dtype=torch.int64))
                instance_masks.append(torch.empty((0, 1, 1), dtype=torch.bool))
        return instance_classes, instance_masks

    def instance_to_panoptic(
        self, scores: Tensor, classes: Tensor, masks: Tensor
    ) -> Tensor:
        # remove low scoring pixels and low scoring instances
        masks = masks * ((masks > 0.5) & rearrange(scores > 0.5, "b i -> b i 1 1"))
        # remove overlaps by keeping the highest scoring instance only at each pixel
        masks = (masks == reduce(masks, "b i h w -> b 1 h w", "max")) & (masks > 0)
        classes = rearrange(classes + self.num_stuff_classes, "b i -> b i 1 1")
        thing_classes = reduce(masks * classes, "b i h w -> b 1 h w", "max")
        batch_size = masks.shape[0]
        all_ids = repeat(torch.arange(self.max_instances), "i -> b i 1 1", b=batch_size)
        all_ids = (all_ids + self.num_stuff_classes).to(masks.device)
        thing_ids = reduce(masks.to(all_ids) * all_ids, "b i h w -> b 1 h w", "max")
        return torch.cat([thing_classes, thing_ids], dim=1)

    def training_step(
        self,
        inputs: List[Tensor],
        targets: Tensor,
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, Any]]:
        # NOTE: targets are concatenated [class_map, id_map]
        # NOTE: category indices are stuff first, then things
        instance_classes, instance_masks = self.panoptic_to_instance(targets)
        instance_loss, instance_metrics = super().training_step(
            inputs, instance_classes, instance_masks, is_validating
        )

        pred_semantic_map = interpolate(
            self.get_semantic_map(inputs), size=targets.shape[2:]
        )
        semantic_target = targets[:, 0]
        is_stuff = semantic_target < self.num_stuff_classes
        ignored = torch.full_like(semantic_target, self.ignore_index) * (~is_stuff)
        semantic_loss = functional.cross_entropy(
            pred_semantic_map,
            semantic_target * is_stuff + ignored,  # ignore thing pixels
            ignore_index=self.ignore_index,
            reduction="mean",
        )

        metrics = dict(**instance_metrics, semantic_loss=semantic_loss)
        return instance_loss + semantic_loss, metrics

    def on_validation_start(self) -> None:
        super().on_validation_start()
        metric_kwargs = {
            "task": "multiclass",
            "num_classes": self.num_stuff_classes + self.num_thing_classes,
            "ignore_index": self.ignore_index,
        }
        self.pixel_accuracy = Accuracy(**metric_kwargs)
        self.mean_iou_computer = JaccardIndex(**metric_kwargs)

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, Any]]:
        num_instances, scores, classes, masks = super().forward(inputs)
        things = self.instance_to_panoptic(scores, classes, masks)
        stuff = interpolate(
            self.get_semantic_map(inputs), size=inputs[0].shape[2:]
        ).argmax(dim=1)
        stuff = stuff * (things[:, 0] == 0)  # zero-out thing pixels

        pred_semantic = stuff + things[:, 0, :, :]
        target_semantic = targets[:, 0, :, :]
        self.mean_iou_computer.to(stuff.device).update(pred_semantic, target_semantic)
        self.pixel_accuracy.to(stuff.device).update(pred_semantic, target_semantic)

        map_computer_preds = []
        for batch_idx, sample_masks in enumerate(masks):
            non_empty_idxs = torch.any(sample_masks > 0.5, dim=(1, 2))
            sample_boxes = masks_to_boxes(sample_masks[non_empty_idxs] > 0.5)
            map_computer_preds.append(
                {
                    "scores": scores[batch_idx][non_empty_idxs],
                    "labels": classes[batch_idx][non_empty_idxs],
                    "boxes": sample_boxes,
                }
            )
        target_classes, target_masks = self.panoptic_to_instance(targets)
        target_boxes = [masks_to_boxes(_) for _ in target_masks]
        self.map_computer.to(stuff.device).update(
            map_computer_preds,
            [
                {"labels": c.to(torch.int64), "boxes": b}
                for c, b in zip(target_classes, target_boxes)
            ],
        )

        loss, metrics = self.training_step(inputs, targets, is_validating=True)
        self.loss_computer.to(loss.device).update(loss)
        return loss, metrics

    def on_validation_end(self) -> Dict[str, float]:
        metrics = super().on_validation_end()
        metrics["pixel_accuracy"] = self.pixel_accuracy.compute().item()
        metrics["mean_iou"] = self.mean_iou_computer.compute().item()
        return metrics



from typing import List

from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import SceneTextRecognition

from .common import get_images, plot_to_numpy


RAINBOW_CMAPS = ["Reds", "Oranges", "Greens", "Blues", "Purples"]


@get_images.register(SceneTextRecognition)
def _(head, config, input, target, features) -> List[np.ndarray]:
    prediction = head(features)
    if "tokens" in config:
        tokens = config["tokens"]
    else:
        tokens = [str(_) for _ in range(head.num_tokens)]

    if isinstance(tokens, str):
        tokens = list(tokens)
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    index_map, token_map = head.get_maps(features)
    index_map, token_map = index_map.to("cpu").numpy(), token_map.to("cpu").numpy()
    visualizations = []

    for batch, image in enumerate(images):
        blank_image = np.full_like(image, fill_value=255)

        fig, axes = plt.subplots(2, 3, figsize=(10.24, 5.12), dpi=100)
        for ax_row in axes:
            for ax in ax_row:
                ax.axis("off")
        axes[0][0].title.set_text("Input")
        axes[0][0].imshow(image)

        axes[0][1].title.set_text("Target")
        axes[0][1].imshow(blank_image)
        if target is not None:
            patch_text = " ".join(
                [tokens[idx] if tokens else f"<{idx}>" for idx in target[batch]]
            )
            set_text(axes[0][1], patch_text.strip() + "\n\n")
        axes[0][2].title.set_text("Prediction")
        axes[0][2].imshow(blank_image)
        if prediction is not None:
            scores = prediction[0][batch].cpu().numpy()
            positive_indices = np.argwhere(scores > 0.5).flatten().tolist()
            if len(positive_indices):
                num_chars = positive_indices[-1] + 1
                pred_tokens = prediction[1][batch][:num_chars]
                scores = scores[:num_chars]

                axes[1][0].title.set_text("Index map")
                for idx, map_slice in enumerate(index_map[batch][:num_chars]):
                    axes[1][0].imshow(
                        map_slice,
                        cmap=RAINBOW_CMAPS[idx % len(RAINBOW_CMAPS)],
                        vmin=0,
                        vmax=1,
                        alpha=map_slice * map_slice.max(),
                        interpolation="none",
                    )
                    set_text_on_peak(axes[1][0], map_slice, text=str(idx))

                axes[1][1].title.set_text("Token map")
                for idx, map_slice in enumerate(token_map[batch]):
                    if idx not in pred_tokens or idx >= len(tokens):
                        continue
                    axes[1][1].imshow(
                        map_slice,
                        cmap=RAINBOW_CMAPS[idx % len(RAINBOW_CMAPS)],
                        vmin=0,
                        vmax=1,
                        alpha=map_slice * map_slice.max(),
                        interpolation="none",
                    )
                    set_text_on_peak(axes[1][1], map_slice, text=tokens[idx])

                patch_text = " ".join(
                    [tokens[idx] for idx in pred_tokens if idx < len(tokens)]
                ).strip()
                patch_text += "\nscores: " + str([round(s, 2) for s in scores.tolist()])
            else:
                patch_text = f"\nmax score: {scores.max()}"

            set_text(axes[0][2], patch_text)

        axes[1][2].imshow(blank_image)
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations


def set_text(ax, text):
    ax.text(
        0.5,
        0.5,
        text,
        transform=ax.transAxes,
        ha="center",
        va="center",
        fontsize="large",
        wrap=True,
    )


def set_text_on_peak(ax, image, text, fontsize=8):
    peak = np.unravel_index(np.argmax(image), image.shape)
    ax.text(
        peak[1],
        peak[0],
        text,
        fontsize=fontsize,
        color="black",
        ha="center",
        va="center",
        bbox={"facecolor": "white", "edgecolor": "none", "alpha": 0.5},
    )



from pathlib import Path
from typing import Tuple, List

from PIL import Image
from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
import kaggle
import lightning
import lightning.pytorch as pl
import logging
import numpy as np
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule
from sihl.heads import DepthEstimation
from sihl.layers import FPN
from sihl.torchvision_backbone import TorchvisionBackbone


class NYU2Dataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, train: bool = False) -> None:
        self.train = train
        self.data_dir = data_dir
        self.min_dist, self.max_dist = 0.0, 10.0
        self.images: List[Tuple[Path, List[float]]] = []
        for imgf in data_dir.rglob("rgb_*.png"):
            depthf = imgf.with_name(imgf.name.replace("rgb", "depth"))
            if not depthf.exists():
                continue
            self.images.append((imgf, depthf))
        if self.train:
            self.transform = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.RandomCrop((512, 640), pad_if_needed=True),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.RandomCrop((512, 640), pad_if_needed=True),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target_path = self.images[idx]
        image = torchvision.io.read_image(str(image_path.resolve()))
        # images are in uint16 format, but we cast to int32 for pytorch compatibility
        target = torch.tensor(np.array(Image.open(target_path)).astype(np.int32))
        image, target = self.transform(tv_tensors.Image(image), tv_tensors.Mask(target))
        mask = target != 0
        target = (target - 1) / (2**16 - 2)
        target = target * (self.max_dist - self.min_dist) + self.min_dist
        return image, {"targets": target, "masks": mask}

    def __len__(self) -> int:
        return len(self.images)


class NYU2DataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "NYU2"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "awsaf49/nyuv2-official-split-dataset",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = NYU2Dataset(self.data_dir / "train", train=True)
        self.valset = NYU2Dataset(self.data_dir / "test", train=False)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="depth_estimation"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_epochs=30,
        accelerator="gpu",
        devices=[0],
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
        gradient_clip_val=1,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True, top_level=7)
        neck = FPN(backbone.out_channels, 256, bottom_level=2, top_level=7)
        head = DepthEstimation(
            in_channels=neck.out_channels,
            lower_bound=0.0,
            upper_bound=10.0,
            bottom_level=2,
            top_level=7,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-4},
            data_config={"range": (0.0, 10.0), "ignore": 0},
        )

    log.debug(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=NYU2DataModule(batch_size=32))



from typing import Tuple, List, Dict

from torch import nn, Tensor
from torchmetrics import MeanMetric
import torch

from sihl.layers import SequentialConvBlocks


class ViewInvarianceLearning(nn.Module):
    """View invariance learning is a kind of self-supervised learning whereby the model
    is encouraged to produce similar representations for different views (i.e.
    distortions) of the same original image.

    Refs:
        1. [Barlow Twins](https://arxiv.org/abs/2103.03230)
    """

    def __init__(
        self,
        in_channels: List[int],
        embedding_dim: int = 1024,
        level: int = 5,
        num_channels: int = 256,
        num_layers: int = 4,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            embedding_dim (int, optional): Number of embedding channels. Defaults to 1024.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 1.
        """
        assert level < len(in_channels)
        assert num_channels > 0 and num_layers > 0
        super().__init__()

        self.in_channels = in_channels
        self.embedding_dim = embedding_dim
        self.level = level
        self.num_channels = num_channels
        self.num_layers = num_layers
        self.projector = nn.Sequential(
            SequentialConvBlocks(in_channels[level], num_channels, num_layers),
            nn.Conv2d(num_channels, embedding_dim, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )
        self.output_shapes = {"representations": ("batch_size", embedding_dim)}

    def forward(self, inputs: List[Tensor]) -> Tensor:
        return self.projector(inputs[self.level])

    def get_correlation(self, inputs1: List[Tensor], inputs2: List[Tensor]) -> Tensor:
        embedding1 = self.projector(inputs1[self.level])
        if embedding1.shape[0] > 1:
            embedding1 = (embedding1 - embedding1.mean(0)) / embedding1.std(0)
        embedding2 = self.projector(inputs2[self.level])
        if embedding2.shape[0] > 1:
            embedding2 = (embedding2 - embedding2.mean(0)) / embedding2.std(0)
        return torch.mm(embedding1.T, embedding2) / embedding1.shape[0]

    def training_step(
        self, inputs1: List[Tensor], inputs2: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        cross_correlation = self.get_correlation(inputs1, inputs2)
        eye = torch.eye(cross_correlation.shape[0], device=cross_correlation.device)
        invariance = (cross_correlation * eye - eye).pow(2).sum()
        redundancy = (cross_correlation * (1 - eye)).pow(2).sum()
        loss = invariance + redundancy / self.num_channels
        return loss, {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.norm_computer = MeanMetric(nan_strategy="ignore")
        self.ondiag_comupter = MeanMetric(nan_strategy="ignore")
        self.offdiag_computer = MeanMetric(nan_strategy="ignore")

    def validation_step(
        self, inputs1: List[Tensor], inputs2: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        loss = self.training_step(inputs1, inputs2)[0]
        device = loss.device
        self.loss_computer.to(device).update(loss)
        cross_correlation = self.get_correlation(inputs1, inputs2).abs()
        dim = cross_correlation.shape[0]
        identity = torch.eye(dim, device=device)
        norm = torch.linalg.matrix_norm(cross_correlation - identity)
        A_norm = torch.linalg.matrix_norm(cross_correlation)
        I_norm = torch.linalg.matrix_norm(identity)
        max_diff_norm = torch.sqrt(A_norm**2 + I_norm**2)
        self.norm_computer.to(device).update(norm / max_diff_norm)
        mask = identity.to(torch.bool)
        self.ondiag_comupter.to(device).update(cross_correlation[mask].mean())
        self.offdiag_computer.to(device).update(cross_correlation[~mask].mean())
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "normalized_frobenius_norm": self.norm_computer.compute().item(),
            "on_diagonal_mean": self.ondiag_comupter.compute().item(),
            "off_diagonal_mean": self.offdiag_computer.compute().item(),
        }



from typing import List, Tuple, Dict

from torch import Tensor, nn
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MeanMetric, R2Score

from sihl.layers import SequentialConvBlocks


class Regression(nn.Module):
    """Regression is the prediction of a scalar within a given finite interval."""

    def __init__(
        self,
        in_channels: List[int],
        lower_bound: float,
        upper_bound: float,
        level: int = 5,
        num_channels: int = 256,
        num_layers: int = 1,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            lower_bound (float): Lower bound of the interval of possible values.
            upper_bound (float): Upper bound of the interval of possible values.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
        """
        assert lower_bound < upper_bound
        assert num_channels > 0 and num_layers > 0
        assert level < len(in_channels)
        super().__init__()

        self.convs = nn.Sequential(
            SequentialConvBlocks(in_channels[level], num_channels, num_layers),
            nn.Conv2d(num_channels, 1, kernel_size=1),
            nn.Sigmoid(),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(0),
        )
        self.level = level
        self.lower_bound = float(lower_bound)
        self.upper_bound = float(upper_bound)
        self.output_shapes = {"values": ("batch_size",)}

    def normalize(self, x: Tensor) -> Tensor:
        return (x - self.lower_bound) / (self.upper_bound - self.lower_bound)

    def denormalize(self, x: Tensor) -> Tensor:
        x = x * (self.upper_bound - self.lower_bound) + self.lower_bound
        return x.clamp(self.lower_bound, self.upper_bound)

    def forward(self, inputs: List[Tensor]) -> Tensor:
        return self.denormalize(self.convs(inputs[self.level]))

    def training_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        pred_values = self.convs(inputs[self.level])
        return (self.normalize(targets) - pred_values).cosh().log().mean(), {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.r2 = R2Score()
        self.mae_computer = MeanAbsoluteError()
        self.mse_computer = MeanSquaredError()

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        pred_values = self.convs(inputs[self.level])
        loss = (self.normalize(targets) - pred_values).cosh().log().mean()
        device = loss.device
        pred_values = self.denormalize(pred_values)
        self.loss_computer.to(device).update(loss)
        self.r2.to(device).update(pred_values, targets)
        self.mae_computer.to(device).update(pred_values, targets)
        self.mse_computer.to(device).update(pred_values, targets)
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "r_squared": self.r2.compute().item(),
            "mean_absolute_error": self.mae_computer.compute().item(),
            "mean_squared_error": self.mse_computer.compute().item(),
        }
